#include <qbutton.h>
#include <qpainter.h>

#include "base.h"
#include "bouton.h"
#include "couleurs.h"

bouton::bouton(const QPixmap & pm1,
			const QPixmap & pm0,
			const char * t, const char * m,
			QWidget * parent)
  : QButton(parent,0)
{
  pointed=false;
  toggled=false;
  justToggled=false;

  touche= new QString(t);
  info  = new QString(m);

  p0     = new QPixmap(pm0);
  p1     = new QPixmap(pm1);
  buffer = new QPixmap;

  setBackgroundColor(C_BE_FOND);
}

bouton::~bouton()
{
  delete buffer;
  delete p0;
  delete p1;
  delete info;
  delete touche;
}

void bouton::enterEvent(QEvent * fe)
{
  pointed=true;
  repaint(false);
  emit texte(*touche,*info);
}

void bouton::leaveEvent(QEvent * fe)
{
  pointed=false;
  repaint(false);
  emit texte(QString(""),QString(""));
}

void bouton::resizeEvent(QResizeEvent * re)
{
  buffer->resize(re->size());
}

void bouton::paintEvent(QPaintEvent * pe)
{
  QPainter p;
  int w=width ()-1;
  int h=height()-1;

  p.begin(buffer);
  p.drawPixmap(0,0,toggled?*p1:*p0,0,0);

#define IMAX (1)
  if (pointed || toggled)
    for (int i=0; i<=IMAX; i++)
	 {
	   int k= 64*i/IMAX;

	   p.setPen(QColor(32+k,32+k,32+k));
	   p.drawLine(w-i,1+i,w-i,h-i);
	   p.drawLine(1+i,h-i,w-i,h-i);
	   
	   p.setPen(QColor(255-k,255-k,255-k));
	   p.drawLine(i,i,w-i,i);
	   p.drawLine(i,i,i,h-i);
	 }
#undef IMAX

  p.end();

  p.begin(this);
  p.drawPixmap(0,0,*buffer);
  p.end();
}

void bouton::mousePressEvent(QMouseEvent * me)
{
  if (me->button()==LeftButton)
    emit clic(this);

  if (me->button()==RightButton)
    emit proprietes(this);
}

void bouton::mouseReleaseEvent(QMouseEvent * me)
{
  if (me->button()==LeftButton && justToggled)
    {
	 justToggled=false;
	 emit relachement(this);
    }
}

bool bouton::estSelectionne()
{
  return toggled;
}

void bouton::selectionne(bool b)
{
  if (b==toggled)
    return;

  toggled=b;
  
  if (b)
    justToggled=true;

  repaint(false);
}
