// -------------------------------------------------------------------------------
// TEST DE VIABILITE:                                                             
// -------------------------------------------------------------------------------
objets gcarte::couturePossible(const brin * b1, const brin * b2) const
{
  if (b1==NULL || b2==NULL)
    return Aucun;
  
  brin * b1a[4], * b2a[4];
  
  for (int i=0; i<4; i++)
    {
	 b1a[i]= b1->getAlpha(i);
	 b2a[i]= b2->getAlpha(i);
    }
  
  if (b1==b2)
    return Aucun;
  
  // Sommets:
  if (b1a[0]==b1 && b2a[0]==b2)
    return Sommets;
  
  if ((b1a[0]==b1 || b2a[0]==b2) && (b1a[0]!=b1 || b2a[0]!=b2))
    return Aucun;
  
  // Aretes:
  if (b1a[1]==b1 && b2a[1]==b2)
    {
	 if (b1a[0]==b2 ||
		b1a[0]->getAlpha(1)==b2a[0]) // Pour viter les cycles de 2 aretes
	   return Aucun;
	 
	 return Aretes;
    }
  
  if ((b1a[1]==b1 || b2a[1]==b2) && (b1a[1]!=b1 || b2a[1]!=b2))
    return Aucun;
  
  // Faces:
  if (b1a[2]==b1 && b2a[2]==b2)
    {
	 if (orbiteFace(b1).contient(b2))
	   return Aucun;
	 else
	   return Faces;
    }
  
  if ((b1a[2]==b1 || b2a[2]==b2) && (b1a[2]!=b1 || b2a[2]!=b2))
    return Aucun;
  
  // Volumes:
  if (b1a[3]==b1 && b2a[3]==b2)
    {
	 if (orbiteFace(b1).taille() != orbiteFace(b2).taille())
	   return Aucun;

	 if (orbiteFace(b1).contient(b2))
	   return Aucun;
	 
	 return Volumes;
    }
  
  return Aucun;
}
// -------------------------------------------------------------------------------
// GEOMETRIE DES COUTURES:                                                        
// -------------------------------------------------------------------------------
void gcarte::orienteObjet(const brin * b1, const brin * b2)
{
  ensemble esc1     = sommets(orbiteComposante(b1));
  sommet   baryFace1= barycentre(sommets(orbiteFace(b1)));
  sommet   baryFace2= barycentre(sommets(orbiteFace(b2)));

  // Normales:
  sommet n1 = normale(b1);
  sommet n2 = normale(b2);
  // Directionelles:
  sommet d1 = (*b1->getSommet() + *b1->getAlpha(0)->getSommet())/2 - baryFace1;
  sommet d2 = (*b2->getSommet() + *b2->getAlpha(0)->getSommet())/2 - baryFace2;
  // Rotations:
  orientation(n1,d1,n2,d2, esc1);
}
//--------------------------------------------------------------------------------
void gcarte::redimensionneObjet(const brin * b1, const brin * b2)
{
  ensemble ES = sommets(orbiteComposante(b1));
  
  coord C1 = circonferenceFace(b1);
  coord C2 = circonferenceFace(b2);
  
  if (!nul(C1))
    agrandissement(sommet(1,1,1)*C2/C1,barycentre(ES),ES);
}
//--------------------------------------------------------------------------------
// Dplacement pour faire concider les barycentres des faces  coudre:
void gcarte::deplaceObjet(const brin * b1, const brin * b2)
{
  ensemble ES= sommets(orbiteComposante(b1));

  translation(
		    barycentre(sommets(orbiteFace(b2)))-
		    barycentre(sommets(orbiteFace(b1))),
		    ES);
}
// -------------------------------------------------------------------------------
// TOPOLOGIE DES COUTURES:                                                        
// -------------------------------------------------------------------------------
bool gcarte::coudre(brin * a, brin * b,
				bool orientation, bool taille, bool position)
{
  switch (couturePossible(a,b))
    {
    case Aucun:
	 {
	   return false;
	 }
    case Sommets:
	 // couture symtrique:
	 {
	   a->setAlpha(0,b);
	   b->setAlpha(0,a);
	   break;
	 }
    case Aretes:
	 // couture asymtrique:
	 {
	   // Plongement:
	   ensemble EB = orbiteComposante(a);

	   if (position && !EB.contient(b))
		{
		  // Deux composantes connexes: on dplace toute la composante connexe de a:
		  ensemble ES= sommets(EB);
		  translation(*b->getSommet() - *a->getSommet() , ES, ES);
		}
	   
	   // Suppression du sommet redondant:
	   a->setSommet(b->getSommet());
	   // Topologie:
	   a->setAlpha(1,b);
	   b->setAlpha(1,a);
	   break;
	 }
    case Faces:
	 // couture asymtrique:
	 {
	   // Suppression des 2 sommets redondants:
	   sommet * s= b->getSommet();
	   ensemble e= orbiteSommet(a);
	   
	   while (!e.estVide())
		((brin *) e.choisis())->setSommet(s);
	   
	   s = b->getAlpha(0)->getSommet();
	   e = orbite(a->getAlpha(0),false,true,true,true);
	   while (!e.estVide())
		((brin *) e.choisis())->setSommet(s);
	   
	   // Topologie:
	   a->setAlpha(2,b);
	   b->setAlpha(2,a);
	   a->getAlpha(0)->setAlpha(2,b->getAlpha(0));
	   b->getAlpha(0)->setAlpha(2,a->getAlpha(0));
	   break;
	 }
    case Volumes:
	 // couture asymtrique:
	 {
	   // Plongement:
	   if (orientation) orienteObjet      (a,b);
	   if (taille     ) redimensionneObjet(a,b);
	   if (position   ) deplaceObjet      (a,b);
	   // Suppression des sommets redondants ET topologie:
	   brin *aa=a, *bb=b; bool alpha0=true; ensemble e;
	   do
		{
		  e= orbiteSommet(aa);
		  while (!e.estVide())
		    ((brin *) e.choisis())->setSommet(bb->getSommet());
		  
		  aa->setAlpha(3,bb);
		  bb->setAlpha(3,aa);
		  
		  if (alpha0)
		    { aa= aa->getAlpha(0); bb= bb->getAlpha(0); }
		  else
		    { aa= aa->getAlpha(1); bb= bb->getAlpha(1); }
		  alpha0=!alpha0;
		}
	   while (a!=aa);
	   break;
	 }
    }
  
  return true;
}
