// -------------------------------------------------------------------------------
// TEST DE VIABILITE:                                                             
// -------------------------------------------------------------------------------
objets gcarte::decouturePossible(const brin * b) const
{
  if (b==NULL)
    return Aucun;

  brin
    * b0   = b  ->getAlpha(0),
    * b01  = b0 ->getAlpha(1),
    * b012 = b01->getAlpha(2),
    * b1   = b  ->getAlpha(1),
    * b12  = b1 ->getAlpha(2),
    * b2   = b  ->getAlpha(2),
    * b23  = b2 ->getAlpha(3),
    * b3   = b  ->getAlpha(3);

  objets o=Volumes;
  if (b3==b) o=Faces  ;
  if (b2==b) o=Aretes ;
  if (b1==b) o=Sommets;
  if (b0==b) o=Aucun  ;
  
  if (o==Aucun                 ) return Aucun;
  // SOMMETS:
  // Dcouture impossible si alpha1(alpha0(b)) est cousu par alpha2.
  // Remarque: ici, b n'est pas cousu par alpha1.
  if (o==Sommets && b01 != b012) return Aucun;
  // ARETES:
  // Dcouture impossible si alpha1(b) est cousu par alpha2.
  // Remarque: ici, b n'est pas cousu par alpha2.
  if (o==Aretes  && b1  != b12 ) return Aucun;
  // FACES:
  // Dcouture impossible si alpha2(b) est cousu par alpha3.
  // Remarque: ici, b n'est pas cousu par alpha3.
  if (o==Faces   && b2  != b23 ) return Aucun;
  // VOLUMES:
  // Dcouture toujours possible.

  return o;
}
// -------------------------------------------------------------------------------
// DECOUTURES:                                                                    
// -------------------------------------------------------------------------------
brin * gcarte::decoudre(brin * b)
{
  brin
    * b0   = b ->getAlpha(0),
    * b01  = b0->getAlpha(1),
    * b02  = b0->getAlpha(2),
    * b1   = b ->getAlpha(1),
    * b2   = b ->getAlpha(2);

  switch (decouturePossible(b))
    {
    case Aucun:
	 {
	   return b;
	 }
    case Sommets: // Dans ce cas: Pas de cration de nouveaux sommets
      {
	   if (b0 == b01)
		b0->setAlpha(0,b0);
	   else
		{
		  b01->setAlpha(1,b01);
		  supprimeBrin(b0);
		}
	   
	   if (b == b1)
		{
		  b->setAlpha(0,b);
		  return b;
		}
	   else
		{
		  supprimeBrin(b);
		  return b1;
		}
      }
    case Aretes:
      {
	   b->setSommet(new sommet(* b->getSommet()));
	   
	   b ->setAlpha(1,b );
	   b1->setAlpha(1,b1);
	   
	   return b;
	 }
    case Faces:
      {
	   // Dcouture topologique:
	   b  ->setAlpha(2,b  );
	   b0 ->setAlpha(2,b0 );
	   b2 ->setAlpha(2,b2 );
	   b02->setAlpha(2,b02);
	   // Duplication ventuelle des sommets:
	   // Il ne faut pas dupliquer le sommet sur lequel le brin b (resp. b2)
	   // pointe si aprs dcouture topologique, alpha2(b) (resp. alpha2(b2))
	   // est contenu dans l'orbite <b,a1,a2,a3> (resp. <b2,a1,a2,a3>).
	   ensemble e;

	   e= orbiteSommet(b);
	   if (!e.contient(b2 ))
		{
		  sommet * ns = new sommet(* b ->getSommet());

		  while (!e.estVide())
		    ((brin *) e.choisis())->setSommet(ns);
		}

	   e = orbiteSommet(b0);
	   if (!e.contient(b02))
		{
		  sommet * ns = new sommet(* b0->getSommet());

		  while (!e.estVide())
		    ((brin *) e.choisis())->setSommet(ns);
		}

	   return b;
      }
    case Volumes:
      {
	   ensemble f = orbiteFace(b);
	   ensemble e;
	   // Dcouture topologique + Duplication ventuelle des sommets:
	   // Pour chaque brin b de la face, il ne faut dupliquer le sommet sur
	   // lequel b pointe que si aprs dcouture topologique, alpha3(b)
	   brin * i = (brin *) f.premier();
	   while (i!=NULL)
		{
		  brin * j = i->getAlpha(3);
		  // Dcouture:
		  i->setAlpha(3,i);
		  j->setAlpha(3,j);
		  // Duplication de sommet:
		  e= orbiteSommet(i);
		  if (!e.contient(j))
		    {
			 sommet * ns = new sommet(* i->getSommet());
			 while (!e.estVide())
			   ((brin *) e.choisis())->setSommet(ns);
		    }
		  // Au suivant!
		  i= (brin *) f.suivant(i);
		}

	   return b;
      }      
    }
}
