#include "base.h"
#include "ensemble.h"

ensemble::ensemble()
{
  creation();
}

ensemble::ensemble(const ensemble & e)
{
  creation();
  operator=(e);
}

void ensemble::creation()
{
  d= new QIntDict<QObject>;
  d->setAutoDelete(false);

  l= new QList<QObject>;
  l->setAutoDelete(false);
}

ensemble::~ensemble()
{
  delete d;
  delete l;
}

long int ensemble::cle(const QObject * o) const
{
  return (long int) o;
}

bool ensemble::contient(const QObject * o) const
{
  return d->find(cle(o)) != NULL;
}

bool ensemble::contientEnsemble(const ensemble & E) const
{
  QObject * o= E.premier();

  while (o!=NULL)
    {
	 if (!contient(o))
	   return false;

	 o= E.suivant(o);
    }

  return true;
}

ensemble & ensemble::enleve(const QObject * o)
{
  if (l->find(o)!=-1)
    {
	 d->take(cle(o));
	 l->take();
    }
  return *this;
}

ensemble & ensemble::ajoute(const QObject * o)
{
  if (o!=NULL && !contient(o))
    {
	 d->insert(cle(o), o);
	 l->append(o);
    }
  return *this;
}

bool ensemble::estVide() const
{
  return l->count()==0;
}

ensemble & ensemble::vide(bool b)
{
  if (b)
    l->setAutoDelete(true);

  l->clear();
  d->clear();
  
  l->setAutoDelete(false);

  return *this;
}

QObject * ensemble::choisis()
{
  if (estVide())
    return NULL;

  QObject * o = l->take(0);
  d->remove(cle(o));

  return o;
}
//-----------------------------------------------------------------
QObject * ensemble::premier() const
{
  return l->first();
}

QObject * ensemble::suivant(const QObject * o) const
{
  if (o!=l->current())
    l->find(o);

  return l->next();
}
//-----------------------------------------------------------------
QObject * ensemble::dernier() const
{
  return l->last();
}

QObject * ensemble::precedant(const QObject * o) const
{
  if (o!=l->current())
    l->find(o);

  return l->prev();
}
//-----------------------------------------------------------------
unsigned int ensemble::taille() const
{
  return l->count();
}

ensemble & ensemble::operator=(const ensemble & e)
{
  if (&e != this)
    {
	 vide();
	 
	 QObject * o = e.premier();
	 
	 while (o!=NULL)
	   {
		ajoute(o);
		o= e.suivant(o);
	   }
    }

  return *this;
}

ensemble & ensemble::enleveEnsemble(const ensemble & e)
{
  QObject * o = e.premier();
  while (o!=NULL)
    {
	 enleve(o);
	 o= e.suivant(o);
    }
  return *this;
}

ensemble & ensemble::ajouteEnsemble(const ensemble & e)
{
  QObject * o = e.premier();
  while (o!=NULL)
    {
	 ajoute(o);
	 o= e.suivant(o);
    }
  return *this;
}
