// Mthodes d'extrusion :      27-31/03/2000 - 02-03/05/2000
// ********************           *************

/* Renvoie 1 si le brin appartient  un sommet seul.  
   2 si le brin appartient  une arte seule.
   3 si le brin appartient  un groupe d'artes.
   4 si le brin appartient  une face seule.
   5 si le brin appartient  un volume. */
int gcarte::app_safv(brin *b)
{
  
  brin *bb;
  ensemble fbrin;
  
  // S'il n'y a qu'un sommet :
  if (b->getAlpha(0)==b)
    return 1;
  
  fbrin = orbiteFace(b);
  
  // S'il n'y a que deux brins, on a une arte.
  if(fbrin.taille()==2)
    return 2;
  
  // Si un brin n'a pas de Alpha 1, on est dans un groupe d'artes.
  b=(brin *) fbrin.premier();
  while (b != NULL)
    {
      if (b->getAlpha(1)==b)
	   return(3);
      b=(brin *) fbrin.suivant(b);
    }
  
  // Si un brin a un Alpha 2, on est dans un volume.
  b=(brin *) fbrin.premier();
  while (b != NULL)
    {
      if (b->getAlpha(2)!=b)
	   return(5);
      b=(brin *) fbrin.suivant(b);
    }
  
  return(4);
}



/* Extrusion selon un vecteur donn :
******************************** */
// v donne les 3 composantes du vecteur d'extrusion.
brin * gcarte::extrusionFace(brin *b, const sommet v)
{
  ensemble fbrin , fsommet, nsommet, nbrin;
  brin *bb;
  int objet=app_safv(b);    // objet auquel appartient b (arte, face ou volume).
  fbrin = orbiteFace(b);    // fbrin est l'ensemble des brins de la face contenant le brin b.
  brin *brin_initial = b;
  
  // Si le brin est dj cousu par alpha 3 : ERREUR !
  if (b->getAlpha(3) != b)
    return b;
  
  // Si on dsire extruder un sommet:
  // *******************************
  if (b->getAlpha(0)==b)
    {
      sommet*s = (sommet*) b->getSommet();      // On cre un nouveau sommet.
      sommet * gerard = new sommet(*s + v);
      
      bb = ajouteBrin(gerard);      // On ajoute un brin au nouveau sommet.
      coudre(b,bb);                 // On le coud au sommet initial.
      return bb ;
    }
  
  
  // Si on dsire extruder une arete :
  // *******************************
  if (objet==2)
    {
      brin *b1, *b2, *b3, *b4, *b5, *b6;
      
      // On rcupre les deux brins de l'arte :
      b = (brin *) fbrin.premier();
      bb= (brin *) fbrin.suivant(b);
      
      // On rcupre les deux sommets de l'arte :
      sommet * s  = (sommet*) b->getSommet();
      sommet * ss = (sommet*) bb->getSommet();
      
      // On cre les deux nouveaux sommets :
      sommet * s1 = new sommet(*s  + v);
      sommet * s2 = new sommet(*ss + v);
      
      // On ajoute deux brins  chaque nouveau sommet et un brin aux anciens sommets :
      b1 = ajouteBrin(s);
      b2 = ajouteBrin(ss);
      b3 = ajouteBrin(s1);
      b4 = ajouteBrin(s1);
      b5 = ajouteBrin(s2);
      b6 = ajouteBrin(s2);
      
      // On relie les brins par les Alpha :
      b1->setAlpha(0,b3);      b3->setAlpha(0,b1);
      b2->setAlpha(0,b6);      b6->setAlpha(0,b2);
      b4->setAlpha(0,b5);      b5->setAlpha(0,b4);
      
      b->setAlpha(1,b1);       b1->setAlpha(1,b);
      bb->setAlpha(1,b2);      b2->setAlpha(1,bb);
      b3->setAlpha(1,b4);      b4->setAlpha(1,b3);
      b5->setAlpha(1,b6);      b6->setAlpha(1,b5);
      
      return b4;
    }
  
  
  // Si on dsire extruder un groupe d'artes :
  // ****************************************
  if (objet==3)
    {
      int nbs = fbrin.taille();     // nb de brins dj existants.
      
      // On remplit l'ensemble fsommet avec les sommets de la face  extruder.
      // --------------------------------------------------------------------
      b = (brin *) fbrin.premier();
      int sortir=0 , i=0;
      // On parcourt les brins de la face.
      while (b!=NULL)
	   {
		// On ajoute le sommet courant  fsommet.
		fsommet.ajoute((sommet*) b->getSommet());
		
		b = (brin *) fbrin.suivant(b);         // On passe au suivant.
		
	   }
      
      
      // Tableau qui contiendra tous les brins de l'ancien groupe de sommets : 48 brins max par sommet.
      brin *tabf[fsommet.taille()+1][48*fsommet.taille()];
      // Tableau qui contiendra tous les brins du nouveau groupe de sommets : 48 brins max par sommet.
      brin *tabn[fsommet.taille()+1][48*fsommet.taille()];      
      
      
      // On remplit le tableau tabf avec les brins du groupe de sommets  extruder.
      // -------------------------------------------------------------------------
      b = (brin *) fbrin.premier();
      sortir=0; i=0;
      while (b != (brin *) fbrin.premier() || sortir == 0)    // On parcourt... 
	   {                                           // ... les brins du groupe.
		tabf[i][0]=b;
		
		b = b->getAlpha(0);           // On prend le brin reli par alpha 0.            
		tabf[i+1][1]=b;
		b = b->getAlpha(1);           // On passe au suivant.
		
		sortir=1;
		++i;
	   }
	 //      tabf[0][1]=tabf[0][0]->getAlpha(1);
      
	 
      
      // On parcourt les sommets de la face et on cre les nouveaux sommets.
      // On en profite pour mettre les nouveaux brins dans nbrin.
      // -------------------------------------------------------
      sommet * s = (sommet*) fsommet.premier();
      i=0;
      while (s!=NULL)
	   {
		sommet * gerard = new sommet(*s + v);
		
		b = ajouteBrin(gerard);      // On ajoute deux brins au nouveau sommet.
		bb = ajouteBrin(gerard);
		tabn[i][0] = bb;             // On ajoute les brins au tableau nbrin.
		tabn[i][1] = b;
		nbrin.ajoute(b);
		nbrin.ajoute(bb);
		
		s = (sommet*) fsommet.suivant(s);
		++i;
	   }
      
      
      // On parcourt les brins du nouveau groupe de sommets.
      // Et on rcupre les nouveaux sommets dans nsommet.
      // ------------------------------------------------
      b = (brin *) nbrin.premier();
      while (b!=NULL)                       
	   { 
		nsommet.ajoute((sommet*) b->getSommet());  
		b = (brin *) nbrin.suivant(b);
	   }
      
      
      /* ************************************************************************
	    On relie les brins du meme sommet par alpha 1 et les brins entre les
	    sommets par alpha 0. Ainsi, on cre un nouveau groupe, copie du
	    premier  une translation prs.
	 ***********************************************************************/
      b = (brin *) nbrin.premier();
      bb = (brin *) nbrin.suivant(b);
      while (bb!=NULL)
	   {
		if (b->getSommet() == bb->getSommet())  // Si les deux brins ...
		  {                               // ... appartiennent au meme sommet :
		    b->setAlpha(1,bb);         //On les relie par Alpha 1
		    bb->setAlpha(1,b);
		  } else {                     // Sinon, on les relie par Alpha 0.
		    b->setAlpha(0,bb);
		    bb->setAlpha(0,b);
		  }
		
		bb = (brin *) nbrin.suivant(bb);       // On passe au brin suivant bb.
		if (bb != (brin *) nbrin.dernier())    // Si ce n'est pas le dernier brin de la liste :
		  {
		    b = (brin *) nbrin.suivant(b);     // On passe b au brin suivant.
		  }
		/*	  else                              // Si bb est le dernier de la liste :
			  {
			  b = (brin *) nbrin.precedant(bb);
			  b->setAlpha(1,bb);
			  bb->setAlpha(1,b);
			  b = (brin *) nbrin.premier();
			  }
		*/	}
	 
      
	 
      
	 // On renvoie le brin image du brin initial :
      for (int i=0 ; i <nsommet.taille() ; i++)
	   for (int j=0 ; j <nbrin.taille()/nsommet.taille() ; j++)
		{
		  if (brin_initial == tabf[i][j])
		    return  tabn[i][j];
		}
      
    }
  
  
  // Si on dsire extruder une face:
  // ******************************
  if (objet==4)
    {      
      int nbs = fbrin.taille();     // nb de brins dj existants.
      
      
      // On remplit l'ensemble fsommet avec les sommets de la face  extruder.
      // --------------------------------------------------------------------
      b = (brin *) fbrin.premier();
      int sortir=0 , i=0;
      // On parcourt les brins de la face.
      while (b != (brin *) fbrin.premier() || sortir == 0)
	   {
		// On ajoute le sommet courant  fsommet.
		fsommet.ajoute((sommet*) b->getSommet());
		
		b = b->getAlpha(0);         // On prend le brin reli par alpha 0.      
		b = b->getAlpha(1);         // On passe au suivant.
		
		sortir=1;      
		
	   }
      
      // Tableau qui contiendra tous les brins de l'ancienne face : 48 brins max par sommet.
      brin *tabf[fsommet.taille()+1][48*fsommet.taille()];
      // Tableau qui contiendra tous les brins de la nouvelle face : 48 brins max par sommet.
      brin *tabn[fsommet.taille()+1][48*fsommet.taille()];
      
	 
      // On remplit le tableau tabf avec les brins de la face  extruder.
      // ---------------------------------------------------------------
      b = (brin *) fbrin.premier();
      sortir=0; i=0;
      while (b != (brin *) fbrin.premier() || sortir == 0)    // On parcourt... 
	   {                                          // ... les brins de la face.
		tabf[i][0]=b;
		
		b = b->getAlpha(0);           // On prend le brin reli par alpha 0.            
		tabf[i+1][1]=b;
		b = b->getAlpha(1);           // On passe au suivant.
		
		sortir=1;
		++i;
	   }
      tabf[0][1]=tabf[0][0]->getAlpha(1);
      
      // On parcourt les sommets de la face et on cre les nouveaux sommets.
      // On en profite pour mettre les nouveaux brins dans nbrin.
      sommet * s = (sommet*) fsommet.premier();
      i=0;
      while (s!=NULL)
	   {
		sommet * gerard = new sommet(*s + v);
		
		b = ajouteBrin(gerard);      // On ajoute deux brins au nouveau sommet.
		bb = ajouteBrin(gerard);
		tabn[i][0] = bb;             // On ajoute les brins au tableau nbrin.
		tabn[i][1] = b;
		nbrin.ajoute(b);
		nbrin.ajoute(bb);
		
		s = (sommet*) fsommet.suivant(s);
		++i;
	   }
	 
      // On parcourt les brins de la nouvelle face.
      // Et on rcupre les nouveaux sommets dans nsommet.
      b = (brin *) nbrin.premier();
      while (b!=NULL)                       
	   { 
		nsommet.ajoute((sommet*) b->getSommet());  
		b = (brin *) nbrin.suivant(b);
	   }
      
      
      /* ************************************************************************
	    On relie les brins du meme sommet par alpha 1 et les brins entre les
	    sommets par alpha 0. Ainsi, on cre une nouvelle face, copie de la
	    premire  une translation prs.
	 ***********************************************************************/
      b = (brin *) nbrin.premier();
      bb = (brin *) nbrin.suivant(b);
      while (bb!=NULL)
	   {
		if (b->getSommet() == bb->getSommet())  // Si les deux brins ...
		  {                               // ... appartiennent au meme sommet :
		    b->setAlpha(1,bb);         //On les relie par Alpha 1
		    bb->setAlpha(1,b);
		  } else {                     // Sinon, on les relie par Alpha 0.
		    b->setAlpha(0,bb);
		    bb->setAlpha(0,b);
		  }
		
		bb = (brin *) nbrin.suivant(bb);       // On passe au brin suivant bb.
		if (bb != (brin *) nbrin.dernier())    // Si ce n'est pas le dernier brin de la liste :
		  {
		    b = (brin *) nbrin.suivant(b);     // On passe b au brin suivant.
		  }
		else                              // Si bb est le dernier de la liste :
		  {
		    b = (brin *) nbrin.precedant(bb);
		    b->setAlpha(1,bb);
		    bb->setAlpha(1,b);
		    b = (brin *) nbrin.premier();
		  }
	   }
      
      
      
      /* ************************************************************************
	    On parcourt les anciens et nouveaux sommets pour leur rajouter 4 brins
	    que l'on rajoute dans les ensembles fbrin et nbrin.
	 ***********************************************************************/
      s = (sommet*) fsommet.premier();
      sommet *ns = (sommet*) nsommet.premier();
      //  while (s != NULL && ns != NULL)
      for (i=0 ; i<nsommet.taille() ; i++)
	   {
		for (int j=2 ; j<6 ; j++)
		  {
		    b =  ajouteBrin(s);
		    bb = ajouteBrin(ns);
		    tabf[i][j] = b;
		    tabn[i][j] = bb;
		    fbrin.ajoute(b);
		    nbrin.ajoute(bb);
		    
		  }
		s  = (sommet*) fsommet.suivant(s) ;
		ns = (sommet*) nsommet.suivant(ns) ;
	   }
      
      for (int i=0 ; i <nsommet.taille() ; i++)
	   {
		// coutures par Alpha 2 :
		// --------------------
		tabf[i][0]->setAlpha(2,tabf[i][3]);
		tabf[i][3]->setAlpha(2,tabf[i][0]);
		tabn[i][0]->setAlpha(2,tabn[i][3]);
		tabn[i][3]->setAlpha(2,tabn[i][0]);
		
		tabf[i][1]->setAlpha(2,tabf[i][2]);
		tabf[i][2]->setAlpha(2,tabf[i][1]);
		tabn[i][1]->setAlpha(2,tabn[i][2]);
		tabn[i][2]->setAlpha(2,tabn[i][1]);
		
		tabf[i][4]->setAlpha(2,tabf[i][5]);
		tabf[i][5]->setAlpha(2,tabf[i][4]);
		tabn[i][4]->setAlpha(2,tabn[i][5]);
		tabn[i][5]->setAlpha(2,tabn[i][4]);
		
		
		// coutures par Alpha 1 :
		// --------------------
		tabf[i][3]->setAlpha(1,tabf[i][4]);
		tabf[i][4]->setAlpha(1,tabf[i][3]);
		tabn[i][3]->setAlpha(1,tabn[i][4]);
		tabn[i][4]->setAlpha(1,tabn[i][3]);
		
		tabf[i][2]->setAlpha(1,tabf[i][5]);
		tabf[i][5]->setAlpha(1,tabf[i][2]);
		tabn[i][2]->setAlpha(1,tabn[i][5]);
		tabn[i][5]->setAlpha(1,tabn[i][2]);
		
		
		// coutures par Alpha 0 :
		// --------------------      
		if (i==nsommet.taille()-1)
		  { 
		    tabf[i][3]->setAlpha(0,tabf[0][2]);
		    tabf[0][2]->setAlpha(0,tabf[i][3]);
		    tabn[i][3]->setAlpha(0,tabn[0][2]);
		    tabn[0][2]->setAlpha(0,tabn[i][3]);
		  } else {
		    tabn[i][3]->setAlpha(0,tabn[i+1][2]);
		    tabn[i+1][2]->setAlpha(0,tabn[i][3]);
		    tabf[i][3]->setAlpha(0,tabf[i+1][2]);
		    tabf[i+1][2]->setAlpha(0,tabf[i][3]);
		  }
		
		tabf[i][4]->setAlpha(0,tabn[i][4]);     // On relie les deux faces.
		tabn[i][4]->setAlpha(0,tabf[i][4]);
		tabf[i][5]->setAlpha(0,tabn[i][5]);
		tabn[i][5]->setAlpha(0,tabf[i][5]);
		
	   }      
      
	 
      // On renvoie le brin image du brin initial :
      for (int i=0 ; i <nsommet.taille() ; i++)
	   for (int j=0 ; j <nbrin.taille()/nsommet.taille() ; j++)
		{
		  if (brin_initial == tabf[i][j])
		    return  tabn[i][j];
		}
	 
    }
  
  
  // Si on dsire extruder un volume :
  // *******************************
  if (objet==5)
    {
      int nbs = fbrin.taille();     // nb de brins dj existants.
      
      
      // On remplit l'ensemble fsommet avec les sommets de la face  extruder.
      // --------------------------------------------------------------------
      b = (brin *) fbrin.premier();
      int sortir=0 , i=0;
      // On parcourt les brins de la face.
      while (b != (brin *) fbrin.premier() || sortir == 0)
	   {
		// On ajoute le sommet courant  fsommet.
		fsommet.ajoute((sommet*) b->getSommet());
		
		b = b->getAlpha(0);         // On prend le sommet reli par alpha 0.      
		b = b->getAlpha(1);         // On passe au suivant.
		
		sortir=1;      
		
	   }
      
      // Tableau qui contiendra tous les brins de l'ancienne face : 48 brins max par sommet.
      brin *tabf[fsommet.taille()+1][48*fsommet.taille()];
      // Tableau qui contiendra tous les brins de la nouvelle face : 48 brins max par sommet.
      brin *tabn[fsommet.taille()+1][48*fsommet.taille()];
      
	 
      // On remplit le tableau tabf avec les brins de la face  extruder.
      // ---------------------------------------------------------------
      b = (brin *) fbrin.premier();
      sortir=0; i=0;
      while (b != (brin *) fbrin.premier() || sortir == 0)    // On parcourt... 
	   {                                          // ... les brins de la face.
		tabf[i][0]=b;
		
		b = b->getAlpha(0);           // On prend le sommet reli par alpha 0.            
		tabf[i+1][1]=b;
		b = b->getAlpha(1);           // On passe au suivant.
		
		sortir=1;
		++i;
	   }
      tabf[0][1]=tabf[0][0]->getAlpha(1);
      
      
      // On parcourt les sommets de la face et on cre les nouveaux sommets.
      // On en profite pour mettre les nouveaux brins dans nbrin.
      // la diffrence avec avant c'est que la nouvelle face se trouve exactement sur l'ancienne.
      sommet * s = (sommet*) fsommet.premier();
      i=0;
      while (s!=NULL)
	   {
		sommet * gerard = new sommet(*s);
		
		b = ajouteBrin(gerard);      // On ajoute deux brins au nouveau sommet.
		bb = ajouteBrin(gerard);
		tabn[i][0] = bb;             // On ajoute les brins au tableau nbrin.
		tabn[i][1] = b;
		nbrin.ajoute(b);
		nbrin.ajoute(bb);
		
		s = (sommet*) fsommet.suivant(s);
		++i;
	   }
      
      // On parcourt les brins de la nouvelle face.
      // Et on rcupre les nouveaux sommets dans nsommet.
      b = (brin *) nbrin.premier();
      while (b!=NULL)                       
	   { 
		nsommet.ajoute((sommet*) b->getSommet());  
		b = (brin *) nbrin.suivant(b);
	   }
	 
      
      /* ************************************************************************
	    On relie les brins du meme sommet par alpha 1 et les brins entre les
	    sommets par alpha 0. Ainsi, on cre une nouvelle face, copie de la
	    seconde  la meme place.
	 ***********************************************************************/
      b = (brin *) nbrin.premier();
      bb = (brin *) nbrin.suivant(b);
      while (bb!=NULL)
	   {
		if (b->getSommet() == bb->getSommet())  // Si les deux brins ...
		  {                               // ... appartiennent au meme sommet :
		    b->setAlpha(1,bb);         //On les relie par Alpha 1
		    bb->setAlpha(1,b);
		  } else {                     // Sinon, on les relie par Alpha 0.
		    b->setAlpha(0,bb);
		    bb->setAlpha(0,b);
		  }
		
		bb = (brin *) nbrin.suivant(bb);       // On passe au brin suivant bb.
		if (bb != (brin *) nbrin.dernier())    // Si ce n'est pas le dernier brin de la liste :
		  {
		    b = (brin *) nbrin.suivant(b);     // On passe b au brin suivant.
		  }
		else                              // Si bb est le dernier de la liste :
		  {
		    b = (brin *) nbrin.precedant(bb);
		    b->setAlpha(1,bb);
		    bb->setAlpha(1,b);
		    b = (brin *) nbrin.premier();
		  }
	   }
      
      b  = tabf[0][0];
      bb = tabn[0][0];
      
      brin * nb = extrusionFace(bb, v);
      coudre(b,bb);
	 
      return  nb;
	 
      
    }
  
}

