#include <qmessagebox.h>
#include <qkeycode.h>
#include <math.h> // Pour la fonction exp
#include "base.h"
#include "fenetre.h"
#include "couleurs.h"
//=====================================================================================================
fenetre::fenetre(QWidget * parent)
  : fenetreGL(parent)
{
  modeActif  = Creation;
  sousMode   = Sommet;
  cg         = true;
  
  antiAlias= false;
  work     = false;
  
  B1=B2=B3=false;
  clicMode=0;
  
  alpha21= true;
  alpha22= true;
  
  cx=cy=cz=true;
  cgo=cgt=cgp=true;
  
  setBackgroundColor(C_FOND);
  setMouseTracking(true);
}

fenetre::~fenetre()
{
}
//-----------------------------------------------------------------------------------
void fenetre::setMode(enum mode m)
{
  if (m==modeActif)
    return;

  if (m==AdjonctionFace || modeActif==AdjonctionFace)
    carte()->setSelection1(NULL,true);

  modeActif=m;
}

enum mode fenetre::getMode()
{
  return modeActif;
}

void fenetre::setSousMode(enum modeObjet m)
{
  sousMode=m;
}

enum modeObjet fenetre::getSousMode()
{
  return sousMode;
}
//-----------------------------------------------------------------------------------
bool fenetre::couturesGeometriques()
{
  return cg;
}

void fenetre::setCouturesGeometriques(bool b)
{
  cg= b;
}
//-----------------------------------------------------------------------------------
bool fenetre::getx() const { return cx; }
bool fenetre::gety() const { return cy; }
bool fenetre::getz() const { return cz; }
void fenetre::setx(const bool b) { cx=b; }
void fenetre::sety(const bool b) { cy=b; }
void fenetre::setz(const bool b) { cz=b; }
bool fenetre::geto() const { return cgo; }
bool fenetre::getp() const { return cgp; }
bool fenetre::gett() const { return cgt; }
void fenetre::seto(const bool b) { cgo=b; }
void fenetre::setp(const bool b) { cgp=b; }
void fenetre::sett(const bool b) { cgt=b; }
//-----------------------------------------------------------------------------------
brin * fenetre::itere(brin * b, bool & alpha2)
{
  brin * resultat;
  
  if (alpha2)
    {
	 resultat= b->getAlpha(2);
	 if (resultat->getAlpha(3)!=resultat)
	   alpha2=false;
    }
  else
    {
	 resultat= b->getAlpha(3);
	 alpha2=true;
    }
  
  return resultat;
}
//=====================================================================================================
sommet fenetre::sommetClique(int x, int y)
{
  brin * b1= carte()->getSelection1();
  sommet
    * s1 = b1->getSommet(),
    * s2 = b1->getAlpha(0)->getSommet();
  
  QPoint
    p1= projection(*s1),
    p2= projection(*s2);
  
  coord t, K= sqr(p2.x()-p1.x()) + sqr(p2.y()-p1.y());
  
  if (nul(K))
    t= 0;
  else
    t= ((p2.x()-p1.x())*(x-p1.x()) + (p2.y()-p1.y())*(y-p1.y()))/K;
  
  if (t<0) t=0;
  if (t>1) t=1;
  
  return *s1 + t*(*s2 - *s1);
}
//=====================================================================================================
void fenetre::beginClic()
{
  brin * b1= carte()->getSelection1();
  brin * bp= carte()->getBrinPointe();
  
  if (!carte()->orbite(bp,false,false,true,true).contient(b1))
    {
	 alpha21= true;
	 b1= bp;
	 carte()->setSelection1(b1);
	 moved=true;
    }
  
  if (bp==NULL)
    return;
  
  // Calcul du sommet cliqu:
  sclic= sommetClique(memoX,memoY);
  QPoint PC = projection(sclic);
  
  memoDX = PC.x() - memoX;
  memoDY = PC.y() - memoY;
  
  // Liste des sommets  dplacer:
  ES.vide();
  memoES.vide();
  ensemble EB;
  
  if (sousMode==Sommet    ) EB = carte()->orbiteSommet    (b1);
  if (sousMode==Arete     ) EB = carte()->orbiteArete     (b1);
  if (sousMode==Face      ) EB = carte()->orbiteFace      (b1);
  if (sousMode==Volume    ) EB = carte()->orbiteVolume    (b1);
  if (sousMode==Composante) EB = carte()->orbiteComposante(b1);
  if (sousMode==Tout      ) EB = carte()->ensembleBrins();
  
  sommet *s;
  while (!EB.estVide())
    {
	 s = ((brin *) EB.choisis())->getSommet();
	 if (!ES.contient(s))
	   {
		ES    .ajoute(s);
		memoES.ajoute(new sommet(*s));
	   }
    }
  
  bary= barycentre(ES);
}

void fenetre::endClic()
{
  ES.vide();
  while (!memoES.estVide())
    delete (sommet *) memoES.choisis();
}
//=====================================================================================================
void fenetre::essaiSelection1(brin * bp)
{
  brin * b1= carte()->getSelection1();
  
  if (carte()->orbite(bp,false,false,true,true).contient(b1))
    changeFace();
  else
    {
	 alpha21= true;
	 carte()->setSelection1(bp);
    }
}

void fenetre::essaiSelection2(brin * bp)
{
  brin * b1= carte()->getSelection1();
  brin * b2= bp;
  
  switch (modeActif)
    {
    case Couture:
	 {	   
	   if (b2!=NULL)
		{
		  ensemble e = carte()->orbite(b2,false,false,true,true);
		  b2= NULL;
		  while (b2==NULL && !e.estVide())
		    {
			 b2 = (brin *) e.choisis();
			 
			 if (!carte()->couturePossible(b1,b2))
			   b2= NULL;
		    }
		}
	   break;
	 }
    case ExtrusionMultiple:
	 {
	   if (!carte()->extrusionMultiplePossible(b1,b2))
		b2=NULL;
	   
	   break;
	 }
    case AdjonctionArete:
	 {
	   ensemble E= carte()->orbite(b2,false,false,true,true);
	   do
		b2= (brin *) E.choisis();
	   while (b2!=NULL && !EB.contient(b2));
	   
	   break;
	 }
    }
  
  alpha22= true;
  carte()->setSelection2(b2);
}
//=====================================================================================================
// GESTION DE LA SOURIS:
//=====================================================================================================
void fenetre::mousePressEvent(QMouseEvent * me)
{
  moved=false;
  work =true ;
  
  bool OldB1= B1;
  bool OldB2= B2;
  bool OldB3= B3;
  
  if (me->button() & LeftButton ) B1=true;
  if (me->button() & MidButton  ) B2=true;
  if (me->button() & RightButton) B3=true;
  
  emit boutons(this,B1,B2,B3);
  
  ensemble BP= brinsProches(me->x(),me->y());
  
  brin * b1= carte()->getSelection1();
  brin * bp= (brin *) BP.premier();
  
  carte()->beginUpdate();
  
  switch (modeActif)
    {
    case Creation:
	 {
	   if (B1 && !B2 && !B3)
		{
		  sommet * s= positionEcranXY(me->x(),me->y(),0);
		  if (s!=NULL)
		    {
			 alpha21= true;
			 carte()->setSelection1(carte()->ajouteBrin(s),true);
		    }
		  else
		    QMessageBox::information(this,"x3gc","Le sommet dsign est trs loin de l'origine.\nIl ne sera pas insr dans le graphe.","OK");
		}

	   if (!B1 && !B2 && B3)
		{
		  brin * b0 = b1;

		  if (b0==NULL || b0->getAlpha(1)!=b1)
		    break;

		  sommet * s= positionEcranXY(me->x(),me->y(),0);
		  if (s==NULL)
		    {
			 QMessageBox::information(this,"x3gc","Le sommet dsign est trs loin de l'origine.\nIl ne sera pas insr dans le graphe.","OK");		  
			 break;
		    }

		  // Cration de l'arte:
		  brin * b2 = carte()->ajouteBrin(s);

		  if (b0->getAlpha(0)!=b0)
		    {
			 brin * b1= carte()->ajouteBrin(b0->getSommet());
			 b1->setAlpha(1,b0);
			 b0->setAlpha(1,b1);
			 
			 b0=b1;
		    }

		  carte()->coudre(b0,b2);
		  // Mise  jour de la slection:
		  alpha21= true;
		  carte()->setSelection1(b2,true);
		}
	   
	   break;
	 }
    case PositionObjet:
    case RotationObjet:
    case TailleObjet:
	 {
	   if (clicMode!=0)
		endClic();
	   
	   clicMode=0;
	   
	   if ((B1 && !B3) || B2) clicMode |= 1;
	   if ((B3 && !B1) || B2) clicMode |= 2;
	   
	   if (clicMode==3 && modeActif==PositionObjet)
		clicMode=0;
	   else
		{
		  memoX= me->x();
		  memoY= me->y();
		  beginClic();
		  
		  if (b1==NULL)
		    clicMode=0;
		}
	   break;
	 }
    case AdjonctionFace:
	 {
	   if (clicMode!=0)
		break;

	   if (B1 && !B2 && !B3)
		{
		  clicMode=1;
		  carte()->addSelection1(bp);
		}

	   if (!B1 && B2 && !B3)
		carte()->setSelection1(NULL,true);
	   
	   
	   if (!B1 && !B2 && B3 && b1!=NULL)
		{
		  ensemble e = carte()->getSelections();
		  carte()->ajouteFace(e);
		}

	   break;
	 }
    case Couture:
    case AdjonctionArete:
    case ExtrusionMultiple:
	 {
	   if (clicMode==0)
		{
		  if (B1 && !B2 && !B3)
		    {
			 clicMode=1;
			 essaiSelection1(bp);
			 
			 if (modeActif==AdjonctionArete)
			   {
				EB= carte()->orbiteFace(carte()->getSelection1());
				
				if (b1!=NULL)
				  {
				    EB.enleveEnsemble(carte()->orbiteSommet(b1                          ));
				    EB.enleveEnsemble(carte()->orbiteSommet(b1             ->getAlpha(0)));
				    EB.enleveEnsemble(carte()->orbiteSommet(b1->getAlpha(1)->getAlpha(0)));
				  }
			   }
		    }
		  
		  if (!B1 && !B2 && B3 && b1!=NULL)
		    {
			 clicMode=2;
			 essaiSelection2(bp);
		    }
		  
		  if (!B1 && B2 && !B3 && modeActif!=ExtrusionMultiple)
		    changeArete();
		}
	   break;
	 }
    case RotationScene:
	 {
   	   if ((B2 && !OldB2) || (B1 && B3 && !OldB1))
		break;
	   
	   if (B1 && !OldB1) clicMode=1;
	   if (B3 && !OldB3) clicMode=2;
	   
	   memoX    = me->x();
	   memoY    = me->y();
	   memoAlpha= getAlpha();
	   memoBeta = getBeta ();
	   memoK    = getK()    ;
	   
	   break;
	 }
    case AdjonctionSommet:
	 {
	   if (B1 && !B2 && !B3)
		{
		  carte()->setSelection1(bp);
		  clicMode=1;
		}
	   break;
	 }
    }
  
  carte()->endUpdate();
}
//=====================================================================================================
void fenetre::mouseReleaseEvent(QMouseEvent * me)
{
  carte()->beginUpdate();
  
  brin * b1= carte()->getSelection1();
  brin * b2= carte()->getSelection2();
  
  bool OldB1= B1;
  bool OldB2= B2;
  bool OldB3= B3;
  
  if (me->button() & LeftButton ) B1=false;
  if (me->button() & MidButton  ) B2=false;
  if (me->button() & RightButton) B3=false;
  
  if (!B1 && !B2 && !B3)
    work=false;
  
  emit boutons(this,B1,B2,B3);
  
  switch (modeActif)
    {
    case RotationScene:
	 {
	   if (clicMode==1 && !B1)
		{
		  clicMode=0;
		  break;
		}
	   
	   if (clicMode==2 && !B3)
		{
		  if (B1)
		    {
			 clicMode=1;
			 memoX    = me->x();
			 memoY    = me->y();
			 memoAlpha= getAlpha();
			 memoBeta = getBeta ();
		    }
		  else
		    clicMode=0;
		}
	   break;
	 }
    case PositionObjet:
    case RotationObjet:
    case TailleObjet:
	 {
	   if (clicMode==0)
		break;
	   
	   if ((clicMode & 1) && !B1 && !B2) clicMode -= 1;
	   if ((clicMode & 2) && !B3 && !B2) clicMode -= 2;
	   
	   if (clicMode==0)
		endClic();
	   
	   if (!moved && carte()->getBrinPointe()!=NULL)
		changeFace();
	   
	   break;
	 }
    case AdjonctionSommet:
	 {
	   if (!B1 && clicMode==1)
		{
		  clicMode=0;
		  
		  if (b1!=NULL)
		    carte()->ajouteSommet(b1,sommetClique(me->x(),me->y()));
		}
	   break;
	 }
    case Couture:
    case AdjonctionArete:
    case AdjonctionFace:
    case ExtrusionMultiple:
	 switch (clicMode)
	   {
	   case 0:
		{
		  break;
		}
	   case 1:
		{
		  if (OldB1 && !B1)
		    clicMode=0;
		  break;
		}
	   case 2:
		{
		  if (OldB3 && !B3)
		    {
			 clicMode=0;
			 
			 if (carte()->getSelection2()!=NULL)
			   switch (modeActif)
				{
				case Couture:
				  {
				    bool geo= cg && !carte()->orbiteComposante(b1).contient(b2);
				    carte()->coudre(b1,b2, geo && cgo, geo && cgt, geo && cgp);
				    break;
				  }
				case AdjonctionArete:
				  {
				    carte()->ajouteArete(b1,b2);
				    break;
				  }
				case ExtrusionMultiple:
				  {
				    carte()->extrusionMultiple(b1,b2);
				    break;
				  }
				}
		    }
		  break;
		}
	   }
    }
  // Pour la mise  jour du brin point:
  enterEvent(NULL);
  
  carte()->endUpdate();
}
//=====================================================================================================
void fenetre::mouseMoveEvent(QMouseEvent * me)
{
  moved= true;
  
  brin * bp;
  
  carte()->beginUpdate();
  
  if (modeActif==Couture || modeActif==AdjonctionSommet || modeActif==AdjonctionArete || modeActif==AdjonctionFace || modeActif==ExtrusionMultiple ||
	 (clicMode==0 && (modeActif==TailleObjet || modeActif==RotationObjet || modeActif==PositionObjet)))
    {
	 bp= (brin *) brinsProches(me->x(),me->y()).premier();
	 
	 if (bp==carte()->getBrinPointe())
	   {
		carte()->endUpdate();
		return;
	   }
	 
	 carte()->setBrinPointe(bp);
    }
  
  switch (modeActif)
    {
    case PositionObjet:
    case RotationObjet:
    case TailleObjet:
	 {
	   if (clicMode==0)
		break;
	   
	   sommet delta;
	   sommet * anc, * nouv;
	   
	   if (clicMode==2)
		{
		  anc  = positionEcranZ(memoX+memoDX,memoY  +memoDY,sclic.getx(),sclic.gety());
		  nouv = positionEcranZ(memoX+memoDX,me->y()+memoDY,sclic.getx(),sclic.gety());
		}
	   else
		{
		  anc  = positionEcranXY(memoX  +memoDX,memoY  +memoDY,sclic.getz());
		  nouv = positionEcranXY(me->x()+memoDX,me->y()+memoDY,sclic.getz());
		}
	   
	   if (nouv==NULL)
		delta = sommet(0,0,0);
	   else
		delta = *nouv - *anc;
	   
	   delete anc;
	   delete nouv;
	   
	   if (modeActif==TailleObjet && clicMode==3)
		{
		  coord k = delta.getx()+delta.gety();
		  delta.setxyz(k,k,k);
		}
	   
	   if (!cx) delta.setx(0);
	   if (!cy) delta.sety(0);
	   if (!cz) delta.setz(0);
	   
	   switch (modeActif)
		{
		case PositionObjet:
		  {
		    translation(delta, memoES,ES);
		    break;
		  }
		case RotationObjet:
		  {
		    delta *= pi/5;
		    
		    if (clicMode==1) { delta.sety(0); delta.setz(0); }
		    if (clicMode==2) { delta.setx(0); delta.sety(0); }
		    if (clicMode==3) { delta.setx(0); delta.setz(0); }
		    
		    rotation(delta,bary, memoES,ES);
		    break;
		  }
		case TailleObjet:
		  {			 
		    delta+= sommet(1,1,1);
		    
		    if (delta.getx()<1) delta.setx(exp(delta.getx()-1));
		    if (delta.gety()<1) delta.sety(exp(delta.gety()-1));
		    if (delta.getz()<1) delta.setz(exp(delta.getz()-1));
		    
		    agrandissement(delta,bary, memoES,ES);
		    break;
		  }
		}
	   break;
	 }
    case AdjonctionSommet:
	 {
	   if (clicMode==1)
		carte()->setSelection1(bp);
	   break;
	 }
    case Couture:
    case AdjonctionArete:
    case ExtrusionMultiple:
	 {
	   switch (clicMode)
		{
		case 1:
		  {
		    alpha21= true;
		    carte()->setSelection1(bp);
		    break;
		  }
		case 2:
		  {
		    essaiSelection2(bp);
		    break;
		  }
		}
	   break;
	 }
    case RotationScene:
	 {
	   if (clicMode==1)
		setAlphaBeta(memoAlpha + (memoX-me->x())/10.0,
				   memoBeta  - (memoY-me->y())/10.0);
	   
	   if (clicMode==2)
		setK(memoK*exp(coord(memoY-me->y())/300));	    
	   
	   break;
	 }
    }
  
  carte()->endUpdate();
}
//=====================================================================================================
void fenetre::changeFace()
{
  if (!((modeActif==Couture                       ) ||
        (modeActif==RotationScene                 ) ||
	   (modeActif==PositionObjet   && clicMode==0) ||
	   (modeActif==RotationObjet   && clicMode==0) ||
	   (modeActif==TailleObjet     && clicMode==0) ||
	   (modeActif==AdjonctionArete               ) ||
	   (modeActif==AdjonctionFace                ) ||
	   (modeActif==ExtrusionMultiple             )))
    return;
  
  brin * b1= carte()->getSelection1();
  brin * b2= carte()->getSelection2();
  
  if (modeActif!=Couture || clicMode!=2)
    {
	 // Modification de la slection 1:
	 if (carte()->orbite(b1,false,false,true,true).contient(carte()->getSelection1()))
	   carte()->setSelection1(itere(b1,alpha21));
    }
  else
    if (b2!=NULL)
	 { 
	   // Slection d'un autre brin destination pour couture!
	   do
		b2= itere(b2,alpha22);
	   while (!carte()->couturePossible(b1,b2));
	   
	   carte()->setSelection2(b2);
	 }
}
//=====================================================================================================
void fenetre::changeArete()
{
  if (!((modeActif==Couture                       ) ||
        (modeActif==RotationScene                 ) ||
	   (modeActif==PositionObjet   && clicMode==0) ||
	   (modeActif==RotationObjet   && clicMode==0) ||
	   (modeActif==TailleObjet     && clicMode==0) ||
	   (modeActif==AdjonctionArete               ) ||
	   (modeActif==ExtrusionMultiple             )))
    return;
  
  bool qui= modeActif==Couture || clicMode!=2;

  brin * b0= qui ? carte()->getSelection1() : carte()->getSelection2();
  brin * b;
  
  if (b0==NULL)
    return;
  
  ensemble F= carte()->orbite(b0,false,false,true,true).enleve(b0);
  ensemble E= carte()->brinsProches(b0).enleveEnsemble(F);

  if (E.taille()==1)
    return;

  if (modeActif==Couture || clicMode!=2)
    {
	 b= (brin *) E.suivant(b0);
	 if (b==NULL)
	   b= (brin *) E.premier();

	 carte()->setSelection1(b);
    }
  else
    {
	 do
	   {
		b= (brin *) E.suivant(b0);
		if (b==NULL)
		  b= (brin *) E.premier();
	   }
	 while (!carte()->couturePossible(carte()->getSelection1(),b));

	 carte()->setSelection2(b);
    }
}
//=====================================================================================================
void fenetre::enterEvent(QEvent * e)
{
QMouseEvent me(Event_MouseMove,mapFromGlobal(QCursor::pos()),0,0);
mouseMoveEvent(&me);
}

void fenetre::leaveEvent(QEvent * e)
{
  carte()->setBrinPointe(NULL);
  DisplayWithAntialias(antiAlias);
}
//=====================================================================================================
bool fenetre::getAntiAlias()
{
  return antiAlias;
}

void fenetre::setAntiAlias(bool b)
{
  if (b==antiAlias)
    return;
  
  antiAlias=b;
  DisplayWithAntialias(antiAlias);
}
//=====================================================================================================
bool fenetre::working()
{
  return work;
}
//=====================================================================================================
