#include <qmessagebox.h>
#include <qpainter.h>
#include <stdlib.h>        // pour la fonction abs!
#include <math.h>          // pour les fonctions trigonomtriques.

#include "erreur.h"
#include "fenetre3D.h"

#define PI ((coord) 3.141592653589793238462643383279)

#define C_SOMMET      (QColor(255,255,255))
#define C_SBRIN       (QColor(255,  0,  0)) // brin slectionn
#define C_SFACE       (QColor(  0,255,  0)) // face slectionne
#define C_SVOLUME     (QColor(255,128,  0)) // volume slectionn
#define C_SCOMPOSANTE (QColor(  0,128,255)) // composante slectionne
#define C_RESTE       (QColor(150,150,150)) // le reste...
#define C_INFOS       (QColor(255,255,255))
#define C_AXES        (QColor(100,100,100))
#define C_GRILLE1     (QColor( 60, 60, 60)) // Couleur de la grille en vue de DESSUS
#define C_GRILLE0     (QColor( 30, 30, 30)) // en vue de DESSOUS (plus foncs)
#define C_FOND        (QColor(  0,  0,  0))

fenetre3D::fenetre3D(QWidget * parent) : QWidget(parent)
{
  SIN= new coord[360];
  COS= new coord[360];
  // Le calcul de sinus et de cosinus est lent.
  // Comme l'on s'en sert sans cesse pour l'affichage 3D, autant les mmoriser.
  // (D'autant plus qu'on travaille en degrs)
  for (int i=0; i<360; i++)
    {
	 COS[i]= cos(PI*i/180);
	 SIN[i]= sin(PI*i/180);
    }
  // Pour les problmes d'arrondis:
  COS[0]=1; COS[90]=0; COS[180]=-1; COS[270]= 0;
  SIN[0]=0; SIN[90]=1; SIN[180]= 0; SIN[270]=-1;

  sbrin      = NULL;
  sface      = new ensemble;
  svolume    = new ensemble;
  scomposante= new ensemble;
  sreste     = new ensemble;

  G= new gcarte;
  buffer= new QPixmap(width(),height());
  bufferOn=true;
  etapesOn=false;

  yoeil=23;
  yplan=17;
  kmul =200;

  alpha=0;
  // beta=  0: vue de dessus,
  // beta= 90: vue de face,
  // beta=180: vue de dessous.
  beta =0;

  modeActif= Creation;

  update=0;
  attenteMAJ=false;

  alpha2= true;
  bbase = NULL;

  setBackgroundColor(C_FOND);

  connect(G,SIGNAL(modifiee(gcarte *)),this,SLOT(carteModifiee(gcarte *)));
}

fenetre3D::~fenetre3D()
{
  delete sreste;
  delete scomposante;
  delete svolume;
  delete sface;
  delete [] SIN;
  delete [] COS;
  delete buffer;
  delete G;
}

gcarte * fenetre3D::carte()
{
  return G;
}

brin * fenetre3D::ajouteBrin(sommet * s)
{
  brin *b= G->ajouteBrin(s);
  maj();
  return b;
}

brin * fenetre3D::getSelection()
{
  return sbrin;
}

int fenetre3D::getAlpha()
{
  return alpha;
}

int fenetre3D::getBeta()
{
  return beta;
}

int fenetre3D::getK()
{
  return (int) kmul;
}

ensemble * fenetre3D::supprimeDoublons(ensemble * e)
{
  ensemble * resultat = new ensemble(e);
  brin * b = (brin *) e->premier();
  
  while (b!=NULL)
    {
	 if (resultat->contient(b))
	   {
		ensemble * f= G->iterateur(b,false,false,true,true);
		f->enleve(b);
		resultat->enleve(f);
	   }
	 b= (brin *) e->suivant(b);
    }

  return resultat;
}

void fenetre3D::setSelection(brin * b)
{
  brin * k;

  *sreste= *(G->ensBrins());
  *scomposante= *(G->composante(b)); 
  *svolume    = *(G->volume    (b));
  *sface      = *(G->face      (b));

  sbrin=b;

  epaisseur   = (G->iterateur(b,false,false,true,true)->taille());

  // puration reste:
  sreste->enleve(scomposante);
  *sreste= *supprimeDoublons(sreste);
  // puration composante:
  k= (brin *) svolume->premier();
  while (k!=NULL)
    {
	 scomposante->enleve(G->iterateur(k,false,false,true,true));
	 k= (brin *) svolume->suivant(k);
    }
  *scomposante= *supprimeDoublons(scomposante);
  // puration volume:
  k= (brin *) sface->premier();
  while (k!=NULL)
    {
	 svolume->enleve(G->iterateur(k,false,false,true,true));
	 k= (brin *) sface->suivant(k);
    }
  *svolume= *supprimeDoublons(svolume);
  // puration face:
  sface->enleve(sbrin);

  maj();
}

void fenetre3D::setAlpha(int a)
{
  a=a%360; if (a<0) a+=360;

  if (a!=alpha)
    {
	 alpha=a;
	 maj();
    }
}

void fenetre3D::setBeta(int b)
{
  if (b<  0) b=  0;
  if (b>180) b=180;

  if (b!=beta)
    {
	 beta=b;
	 maj();
    }
}

void fenetre3D::setK(int k)
{
  if (k<  30) k=  30;
  if (k>1000) k=1000;

  if (k!=kmul)
    {
	 kmul=k;
	 maj();
    }
}

enum mode fenetre3D::getMode()
{
  return modeActif;
}

void fenetre3D::setMode(enum mode m)
{
  modeActif= m;
}

bool fenetre3D::doubleBuffer()
{
  return bufferOn;
}

void fenetre3D::setDoubleBuffer(bool b)
{
  bufferOn=b;
}

bool fenetre3D::etapes()
{
  return etapesOn;
}

void fenetre3D::setEtapes(bool b)
{
  etapesOn=b;
}

void fenetre3D::beginUpdate()
{
  update++;
}

void fenetre3D::endUpdate()
{
  update--;
  
  if (update<0)
    erreur("fenetre3D: endUpdate() appele avant beginUpdate...");

  if (update==0 && attenteMAJ)
    {
	 attenteMAJ=false;
	 majBuffer();
    }
}

void fenetre3D::maj()
{
  if (update>0)
    attenteMAJ= true;
  else
    majBuffer();
}

void fenetre3D::resizeEvent(QResizeEvent * re)
{
  buffer->resize(re->size());
  majBuffer();
}
//=====================================================================================================
QIntDict<QPoint> * fenetre3D::plongement()
{
  QIntDict<QPoint> * E = new QIntDict<QPoint>;
  E->setAutoDelete(true);

  sommet *s= G->sommet1(), *sr;
  while (s!=NULL)
    {
	 sr= rotation(s);
	 // La cl de l'objet est son adresse (transforme en nombre entier)
	 E->insert((long) s,new QPoint(xp(sr),yp(sr)));

	 s= G->sommetSuivant(s);
    }
  return E;
}
//=====================================================================================================
// GESTION DE LA SOURIS:
//=====================================================================================================
void fenetre3D::mousePressEvent(QMouseEvent * me)
{
  if (me->button() & LeftButton ) B1=true;
  if (me->button() & MidButton  ) B2=true;
  if (me->button() & RightButton) B3=true;

  if (!etapes()) beginUpdate();

  switch (modeActif)
    {
    case Creation:
	 {
	   if (B1 && !B2 && !B3)
		{
		  sommet * s= positionBrin(me->x(),me->y());
		  if (s)
		    setSelection( G->ajouteBrin(s) );
		  else
		    QMessageBox::information(this,"x3gc","Le sommet dsign est trs loin de l'origine.\nIl ne sera pas insr dans le graphe.","OK");
		}		  
	   
	   break;
	 }
    case Couture:
	 {
	   if ((!B1 || B2 || B3) && (B1 || B2 || !B3))
		break;

	   if (B1)
		{
		  brin * p= brinProche(me->x(),me->y());

		  if (p!=NULL && p==bbase)
		    {
			 if (alpha2)
			   {
				p= sbrin->getAlpha(2);
				if (p->getAlpha(3)!=p)
				  alpha2=false;
			   }
			 else
			   {
				p= sbrin->getAlpha(3);
				alpha2=true;
			   }
		    }
		  else
		    bbase=p;
		  
		  setSelection(p);
		}
	   else
		{
		  // couture!
		  brin * b1= getSelection();
		  brin * b2= brinProche(me->x(),me->y());

		  if (b1!=NULL)
		    {
			 if (b1==b2)
			   break;
			 
			 if (b1->getAlpha(3)!=b1 || b2->getAlpha(3)!=b2)
			   {
				QMessageBox::information(this,"x3gc","Ce brin est satur.\nVous ne pouvez pas le relier  un autre.","OK");
				break;
			   }
			 
			 if (b1->getAlpha(0)==b1 && b2->getAlpha(0)==b2) G->coudre(SOMMETS,b1,b2); else
			   if (b1->getAlpha(1)==b1 && b2->getAlpha(1)==b2) G->coudre(ARETES ,b1,b2); else
				if (b1->getAlpha(2)==b1 && b2->getAlpha(2)==b2) G->coudre(FACES  ,b1,b2); else
				  if (b1->getAlpha(3)==b1 && b2->getAlpha(3)==b2) G->coudre(VOLUMES,b1,b2);
			 
			 setSelection(getSelection());
		    }
		}
	   break;
	 }
    case ZoomScene:
    case RotationScene:
	 {
	   memoX    = me->globalX();
	   memoY    = me->globalY();
	   memoAlpha= alpha;
	   memoBeta = beta ;
	   memoK    = kmul ;
	   break;
	 }
    }

  if (!etapes()) endUpdate();
}

void fenetre3D::mouseReleaseEvent(QMouseEvent * me)
{
  if (me->button() & LeftButton ) B1=false;
  if (me->button() & MidButton  ) B2=false;
  if (me->button() & RightButton) B3=false;
}

void fenetre3D::mouseMoveEvent   (QMouseEvent * me)
{
  int
    x= me->globalX(),
    y= me->globalY();

  switch (modeActif)
    {
    case ZoomScene:
	 {
	   if (B1 || B2 || B3)
		setK(memoK*exp(coord(memoY-y + x-memoX)/100));
	   break;
	 }
    case RotationScene:
	 {
	   beginUpdate();
	   if (B1 || B2) setBeta (memoBeta  - (memoY-y)/2);
	   if (B2 || B3) setAlpha(memoAlpha + (memoX-x)/2);
	   endUpdate();
	   break;
	 }
    }
}
//=====================================================================================================
// Conversion des coordonnes:
sommet * fenetre3D::rotation(sommet * s)
{
  coord x ,y ,z ;
  coord xa,ya,za;
  coord xb,yb,zb;

  x= s->getx();
  y= s->gety();
  z= s->getz();

  // premire rotation:
  xa= x*COS[alpha] - y*SIN[alpha];
  ya= x*SIN[alpha] + y*COS[alpha];
  za= z;

  // deuxime rotation:
  xb= xa;
  yb= ya*SIN[beta] - za*COS[beta];
  zb= ya*COS[beta] + za*SIN[beta];

  return new sommet(xb,yb,zb);
}
//-----------------------------------------------------------------------------------------------------
// projections sur l'cran: (pour l'affichage)
int fenetre3D::xp(sommet * s)
{
  return
    int (
	    width()/2 +
	    kmul * ( s->getx() * (yoeil-yplan) / (yoeil+s->gety()) ));
}

int fenetre3D::yp(sommet * s)
{
  return
    int (
	    height()/2 -
	    kmul * ( s->getz() * (yoeil-yplan) / (yoeil+s->gety()) ));
}
// projections sur le plan horizontal (z=0) (pour la cration de brins)
#define ABS(x) ((x)<0 ? -(x) : (x))
sommet * fenetre3D::positionBrin(int x0, int z0)
{
  if (beta==90)
    return NULL;

  coord x1,z1,x2,y2,x3,y3;
  // coordonnes du point cliqu  l'cran:
  x1 = +coord(x0-width ()/2)/kmul*yoeil/(yoeil-yplan);
  z1 = -coord(z0-height()/2)/kmul*yoeil/(yoeil-yplan);
  // rotation du plan z=0 de -beta sur Ox et calcul du projet du point cliqu:
  // beta!=90, et beta est dans [0..180] donc: cosinus non nul.
  coord k= yoeil*COS[beta]-z1*SIN[beta];
  if (ABS(k)<1E-5)
    return NULL;

  x2 = x1*(1+z1*SIN[beta]/k);
  y2 = z1*yoeil/k;
  // rotation de -apha sur Oz:
  x3= x2*COS[alpha] + y2*SIN[alpha];
  y3= -x2*SIN[alpha] + y2*COS[alpha];

  if (ABS(x3)>yplan || ABS(y3)>yplan)
    return NULL;

  return new sommet(x3,y3,0);
}
//-----------------------------------------------------------------------------------------------------
int   sqr(int   x) { return x*x; }
coord sqr(coord x) { return x*x; }

int fenetre3D::distance(int x,int y,int x1, int y1, int x2, int y2)
{
  // Distance  un point:
  if (x1==x2 && y1==y2)
    return sqr(x-x1)+sqr(y-y1);

  coord beta = - coord( (x1-x2)*(x-x1) + (y-y1)*(y1-y2))/ ( sqr(x1-x2)+sqr(y1-y2) );
  // On est en dehors de la bande et du cot du point (x1,y1):
  if (beta<=0) return sqr(x-x1) + sqr(y-y1);
  // On est en dehors de la bande et du cot du point (x2,y2):
  if (beta>=1) return sqr(x-x2) + sqr(y-y2);
  // On est dans la bande du segment, il faut calculer la distance  la droite:
  return int( sqr(beta*(x2-x1)+x1-x) + sqr(beta*(y2-y1)+y1-y) );
}

brin * fenetre3D::brinProche(int x,int y)
{
  QIntDict<QPoint> * table= plongement();
  QPoint p1,p2;

  brin * b  = G->brin1();
  brin * sol= NULL;
  int d= 99999, nd;
  
  while (b!=NULL)
    {
	 p1= * (*table)[(long) b->             getSommet()];
	 p2= * (*table)[(long) b->getAlpha(0)->getSommet()];
	 p2= (p1+p2)/2;

	 nd= distance( x,y, p1.x(),p1.y(), p2.x(),p2.y() );

	 if (nd<d)
	   {
		d=nd;
		sol=b;
	   }
	 
	 b= G->brinSuivant(b);
    }
  return sol;
}
#undef ABS(x)
//-----------------------------------------------------------------------------------------------------
void fenetre3D::carteModifiee(gcarte * g)
{
  maj();
}
//=====================================================================================================
// AFFICHAGE 3D DE LA 3G-CARTE:                                                                        
//=====================================================================================================
#define LIGNE(s1,s2) (p->drawLine(xp(s1),yp(s1),xp(s2),yp(s2)))
#define N (5)
//..............................................
void fenetre3D::dessineQuadrillage(QPainter * p)
{
  sommet *s1, *s2;
  p->setPen(beta<=90 ? C_GRILLE1 : C_GRILLE0);
  for (int i=-N; i<=N; i++)
    if (i!=0)
	 {
	   s1= rotation(new sommet(-N, i,0));
	   s2= rotation(new sommet(+N, i,0));
	   LIGNE(s1,s2);
	   s1= rotation(new sommet( i,-N,0));
	   s2= rotation(new sommet( i,+N,0));
	   LIGNE(s1,s2);
	 }
  s1= rotation(new sommet(-N, 0,0));
  s2= rotation(new sommet( 0, 0,0)); LIGNE(s1,s2);
  s1= rotation(new sommet( 1, 0,0));
  s2= rotation(new sommet(+N, 0,0)); LIGNE(s1,s2);

  s1= rotation(new sommet( 0,-N,0));
  s2= rotation(new sommet( 0, 0,0)); LIGNE(s1,s2);
  s1= rotation(new sommet( 0, 1,0));
  s2= rotation(new sommet( 0,+N,0)); LIGNE(s1,s2);
}
#undef  N

#define T (6)
#define R (2*T+1)
#define D (int (6/sqrt(2)))
//..................................................................................
void fenetre3D::dessineNormale(QPainter * p, char * etiquette,
						 sommet * s, int a1, int b1, int a2, int b2)
{
  sommet *O= rotation(new sommet(0,0,0));
  sommet *r= rotation(s);
  int ox= xp(O), oy=yp(O);
  int rx= xp(r), ry=yp(r);
  p->setPen(C_AXES);
  p->drawLine(ox,oy,rx,ry);

  coord K= 1+40/coord(getK());
  sommet * q= rotation(new sommet(K*s->getx(),K*s->gety(),K*s->getz()));
  int dx=0,dy=0;

  if (((alpha==a1 || a1==-1) && beta==b1) || ((alpha==a2 || a2==-1) && beta==b2))
    {
	 dx=10;dy=-10;
  	 p->drawEllipse(ox-T,oy-T,R,R);
	 if ((alpha==a2 && a2!=-1) || (beta==b2 && a2==-1))
	   {
		p->drawLine(ox-D,oy-D,ox+D,oy+D);
		p->drawLine(ox-D,oy+D,ox+D,oy-D);
	   }
    }
  p->drawText(xp(q)+dx-10,yp(q)+dy-10,21,21,AlignCenter,etiquette);
}
#undef R
#undef T
//.............................................................................................
void fenetre3D::dessineEnsemble(QPainter * p, ensemble * e, QColor c, QIntDict<QPoint> * table)
{
  QPoint pos1,pos2;

  p->setPen(c);
  brin * b= (brin *) e->premier();
  while (b!=NULL)
    {
	 pos1= * (*table)[(long) b->getSommet()];
		
	 if (b->getAlpha(0)==b)
	   {
		// sommet isol:
		if (b==getSelection())
		  p->drawEllipse(pos1.x()-3,pos1.y()-3,7,7);
	   }
	 else
	   {
		// arete:
		pos2= * (*table)[(long) b->getAlpha(0)->getSommet()];
		pos2= (pos1+pos2)/2;
		p->drawLine(pos1.x(),pos1.y(),pos2.x(),pos2.y());
	   }
	 
	 b= (brin *) e->suivant(b);
    }
}
//.........................................
void fenetre3D::dessineGraphe(QPainter * p)
{
  QIntDict<QPoint> * table= plongement();

  // Affichage des brins:
  ensemble e;
  e.insere(getSelection());

  dessineEnsemble(p,sreste     ,C_RESTE      ,table);
  dessineEnsemble(p,scomposante,C_SCOMPOSANTE,table);
  dessineEnsemble(p,svolume    ,C_SVOLUME    ,table);
  dessineEnsemble(p,sface      ,C_SFACE      ,table);
  dessineEnsemble(p,&e         ,C_SBRIN      ,table);

  // Affichage des sommets:
  QPoint pos;
  sommet *s= G->sommet1();
  p->setPen(C_SOMMET);
  while (s!=NULL)
    {
	 pos= * (*table)[(long) s];
	 
	 p->drawLine(pos.x()-1,pos.y()  ,pos.x()+1,pos.y()  );
	 p->drawLine(pos.x()  ,pos.y()-1,pos.x()  ,pos.y()+1);
	 
	 s= G->sommetSuivant(s);
    }
}
#undef LIGNE
//........................................
void fenetre3D::dessineInfos(QPainter * p)
{
  QString sb; sb.sprintf("< Brins  : %4d >"  ,G->nbBrins  ());
  QString ss; ss.sprintf("< Sommets: %4d >"  ,G->nbSommets());
  QString se; if (sbrin!=NULL) se.sprintf("< Epaisseur: %2d >",epaisseur);

  p->setFont(QFont("Fixed",10));
  p->setPen(C_INFOS);

  int h= p->fontMetrics().height();
  int H= height()-2*h;

  p->drawText(10,H+0*h,sb);
  p->drawText(10,H+1*h,ss);
  
  if (sbrin!=NULL)
    p->drawText(width()-10-p->fontMetrics().width(se),H+1*h,se);
}
//.........................
void fenetre3D::majBuffer()
{
  QPainter p;
  if (bufferOn)
    p.begin(buffer);
  else
    p.begin(this);

  p.fillRect(0,0,width(),height(),backgroundColor());

  if (beta>=90) dessineNormale(&p,"Z",new sommet(0,0,1),-1,0,-1,180);
  dessineQuadrillage(&p);
  if (alpha>0 && alpha<=180)
    {
	 dessineNormale(&p,"X",new sommet(1,0,0),270,90,90,90);
	 dessineNormale(&p,"Y",new sommet(0,1,0),180,90, 0,90);
    }
  else
    {
	 dessineNormale(&p,"Y",new sommet(0,1,0),180,90, 0,90);
	 dessineNormale(&p,"X",new sommet(1,0,0),270,90,90,90);
    }
  if (beta<90) dessineNormale(&p,"Z",new sommet(0,0,1),-1,0,-1,180);
  dessineGraphe(&p);
  dessineInfos(&p);

  p.end();

  if (bufferOn)
    repaint(false);
}

void fenetre3D::paintEvent(QPaintEvent * pe)
{
  QPainter p;

  p.begin(this);
  p.drawPixmap(0,0,*buffer);
  p.end();
}
//======================================================================================================
