#include <qmessagebox.h>
#include <qpainter.h>
#include <stdlib.h>        // pour la fonction abs!
#include <math.h>          // pour les fonctions trigonomtriques.

#include "base.h"
#include "fenetreGL.h"
#include "accpersp.h"
#include "couleurs.h"

#include <stdlib.h>
#include <stdio.h>

extern "C" void tShade(int,GLdouble*,int);

float xV,yV,zV;

fenetreGL::fenetreGL(QWidget *parent)
  : QGLWidget(parent)
{
  QGLFormat f;
  f.setAccum(true);
  setFormat(f);
  
  ViewWidth = 1;
  
  couleurs= new QList<QColor>;
  couleurs->append(new QColor(C_SBRIN      ));
  couleurs->append(new QColor(C_SBRIN2     ));
  couleurs->append(new QColor(C_SFACE      ));
  couleurs->append(new QColor(C_SFACE2     ));
  couleurs->append(new QColor(C_SVOLUME    ));
  couleurs->append(new QColor(C_SCOMPOSANTE));
  couleurs->append(new QColor(C_RESTE      ));
  
  G= new superCarte();
  alpha=225;
  beta =-45;
  kmul=1;
  rayon = 10;
  xV = 0;yV = 0;zV = 0;
  Shade = true;
  Blend = true;
  Eclate = false;
  Lighting = true;
  
  // private
  Antialias = false;
  
  connect(G   ,SIGNAL(carteModifiee(superCarte *)),
		this,SLOT  (carteModifiee(superCarte *)));		
}

fenetreGL::~fenetreGL()
{
  delete G;
}

superCarte * fenetreGL::carte()
{
  return G;
}

void fenetreGL::maj()
{
//  majBuffer();
	repaint(false);
}

// projections sur le plan horizontal (z=0) (pour la cration de brins)
double view[32*20],proj[32*20];
int vp[32*20];

sommet *fenetreGL::positionEcranZ(int x0,int y0, coord x, coord y)
{
  sommet *s = positionEcranYZ(x0,y0,x);
  if (s==NULL) s = positionEcranXZ(x0,y0,y);
  return new sommet(x,y,s->getz());
}

sommet *fenetreGL::positionEcranXY(int x0, int y0, coord toz, bool borde)
{
  double 	x,y,z,
    x1,y1,z1,
    x2,y2,z2;
  
  glGetDoublev(GL_MODELVIEW_MATRIX,view);
  glGetDoublev(GL_PROJECTION_MATRIX,proj);
  glGetIntegerv(GL_VIEWPORT,vp);
  
  gluUnProject(x0,ViewportHeight-y0,0,view,proj,vp,
			&x1,&y1,&z1);
  gluUnProject(x0,ViewportHeight-y0,1,view,proj,vp,
			&x2,&y2,&z2);
  
  if (z2==z1) return NULL;
  
  x = x1 + ((x2-x1)/(z2-z1) * (toz-z1));
  y = y1 + ((y2-y1)/(z2-z1) * (toz-z1));
  z = toz;
  
  return new sommet(x,y,z);
}

sommet * fenetreGL::positionEcranXZ(int x0, int y0, coord toy, bool borde)
{
  double 	x,y,z,
    x1,y1,z1,
    x2,y2,z2;
  
  glGetDoublev(GL_MODELVIEW_MATRIX,view);	
  glGetDoublev(GL_PROJECTION_MATRIX,proj);
  glGetIntegerv(GL_VIEWPORT,vp);
  
  gluUnProject(x0,ViewportHeight-y0,0,view,proj,vp,
			&x1,&y1,&z1);
  gluUnProject(x0,ViewportHeight-y0,1,view,proj,vp,
			&x2,&y2,&z2);
  
  if (y2==y1) return NULL;
  
  x = x1 + ((x2-x1)/(y2-y1) * (toy-y1));
  y = toy;
  z = z1 + ((z2-z1)/(y2-y1) * (toy-y1));
  
  return new sommet(x,y,z);
}

sommet * fenetreGL::positionEcranYZ(int x0, int y0, coord tox, bool borde)
{
  double 	x,y,z,
    x1,y1,z1,
    x2,y2,z2;
  
  glGetDoublev(GL_MODELVIEW_MATRIX,view);	
  glGetDoublev(GL_PROJECTION_MATRIX,proj);
  glGetIntegerv(GL_VIEWPORT,vp);
  
  gluUnProject(x0,ViewportHeight-y0,0,view,proj,vp,
			&x1,&y1,&z1);
  gluUnProject(x0,ViewportHeight-y0,1,view,proj,vp,
			&x2,&y2,&z2);
  
  if (x2==x1) return NULL;
  
  x =tox;
  y = y1 + ((y2-y1)/(x2-x1) * (tox-x1));
  z = z1 + ((z2-z1)/(x2-x1) * (tox-x1));
  
  return new sommet(x,y,z);
}

QPoint fenetreGL::projection(sommet s)
{
  double 	x,y,z;
  glGetDoublev(GL_MODELVIEW_MATRIX,view);	
  glGetDoublev(GL_PROJECTION_MATRIX,proj);
  glGetIntegerv(GL_VIEWPORT,vp);
  
  gluProject(s.getx(),s.gety(),s.getz(),view,proj,vp,
		   &x,&y,&z);
  
  return QPoint(x,ViewportHeight-y);
}

//-----------------------------------------------------------------------------------------------------

int fenetreGL::distance(int x,int y,int x1, int y1, int x2, int y2)
{
  // Distance  un point:
  if (x1==x2 && y1==y2)
    return sqr(x-x1)+sqr(y-y1);
  
  coord beta = - coord( (x1-x2)*(x-x1) + (y-y1)*(y1-y2))/ ( sqr(x1-x2)+sqr(y1-y2) );
  // On est en dehors de la bande et du cot du point (x1,y1):
  if (beta<=0) return sqr(x-x1) + sqr(y-y1);
  // On est en dehors de la bande et du cot du point (x2,y2):
  if (beta>=1) return sqr(x-x2) + sqr(y-y2);
  // On est dans la bande du segment, il faut calculer la distance  la droite:
  return int( sqr(beta*(x2-x1)+x1-x) + sqr(beta*(y2-y1)+y1-y) );
}

#define DMAX (30)
ensemble fenetreGL::brinsProches(int x,int y)
{
  ensemble resultat;
  unsigned int d= DMAX*DMAX, nd;
  
  QPoint p1,p2;  
  double xx,yy,zz;
  
  glGetDoublev(GL_MODELVIEW_MATRIX,view);	
  glGetDoublev(GL_PROJECTION_MATRIX,proj);
  glGetIntegerv(GL_VIEWPORT,vp);	
  
  sommet * s;
  
  brin * b= G->brin1();
  while (b!=NULL)
    {
	 s= b->getSommet();
	 gluProject(s->getx(),s->gety(),s->getz(), view,proj,vp, &xx,&yy,&zz);
	 p1= QPoint( xx , ViewportHeight-yy );
	 
	 s= b->getAlpha(0)->getSommet();
	 gluProject(s->getx(),s->gety(),s->getz(), view,proj,vp, &xx,&yy,&zz);
	 p2= QPoint( xx , ViewportHeight-yy );
	 
	 p2= (p1+p2)/2;
	 nd= distance( x,y, p1.x(),p1.y(), p2.x(),p2.y() );
	 
	 if (nd<=d)
	   {
		if (nd<d)
		  {
		    resultat.vide();
		    d=nd;
		  }
		resultat.ajoute(b);
	   }
	 
	 b= G->brinSuivant(b);
    }
  
  return resultat;
}
//-----------------------------------------------------------------------------------------------------
void fenetreGL::carteModifiee(superCarte * g)
{
  majBuffer();
}
//=====================================================================================================
// AFFICHAGE 3D DE LA 3G-CARTE:                                                                        
//=====================================================================================================
#define LIGNE(s1,s2) (p->drawLine(xp(s1),yp(s1),xp(s2),yp(s2)))
#define N (5)
//..............................................
void fenetreGL::dessineQuadrillage()
{
  float	QuadMinX = -30,
    QuadMaxX = 30,
    QuadMinY = -30,
    QuadMaxY = 30,
    QuadStepX = 1.0,
    QuadStepY = 1.0;
  float x,y;
  
  glNewList(2,GL_COMPILE);
  
/*	glEnable(GL_FOG);
	glFogi(GL_FOG_MODE,GL_LINEAR);
	glFogf(GL_FOG_START,0);
	glFogf(GL_FOG_END,50);
//	glFogf(GL_FOG_DENSITY,0.01);*/
  
  
  glBegin(GL_LINES);
  glColor3f(0.4,0.4,0.4);
  for (x=QuadMinX;x<=QuadMaxX;x+=QuadStepX) {
    if (x!=0) {
	 glVertex3f(x,QuadMinY,0);
	 glVertex3f(x,QuadMaxY,0); }}
  for (y=QuadMinY;y<=QuadMaxY;y+=QuadStepY) {
    if (y!=0) {
	 glVertex3f(QuadMinX,y,0);
	 glVertex3f(QuadMaxX,y,0); }}
  
  glVertex3f(QuadMinX,0,0);glVertex3f(0,0,0);
  glVertex3f(1,0,0);glVertex3f(QuadMaxX,0,0);	
  glVertex3f(0,QuadMinY,0);glVertex3f(0,0,0);
  glVertex3f(0,1,0);glVertex3f(0,QuadMaxY,0);	
  
  glColor3f(0.5,0.5,1.0);		
  glVertex3f(0,0,0);glVertex3f(1,0,0);
  glVertex3f(0,0,0);glVertex3f(0,1,0);
  glVertex3f(0,0,0);glVertex3f(0,0,1);
  glEnd();
//	glDisable(GL_FOG);
  glEndList();
}
#undef  N
//.............................................................................................
#define R (4)
void fenetreGL::dessineEnsemble(ensemble * e, ensemble * pointes, QColor c)//, QIntDict<QPoint> * table)
{
//  QPoint pos1,pos2;
  
//  p->setPen(c);
  brin * b= (brin *) e->premier();
  while (b!=NULL)
    {
//	 pos1= * (*table)[(long) b->getSommet()];
	 
//	 QPen pen= p->pen();
//	 pen.setWidth(pointes->contient(b) ? 3 : 0);
	 glLineWidth(pointes->contient(b) ? 3 : 1);
//	 p->setPen(pen);
	 
	 sommet *s = b->getSommet();	
	 if (b->getAlpha(0)==b)
	   {
		// sommet isol:
		if (b==G->getSelection1() || b==G->getSelection2() || pointes->contient(b))
//		  p->drawRect(pos1.x()-R,pos1.y()-R,2*R+1,2*R+1);
		  glBegin(GL_POINTS);
		glColor3f(((float)c.red())/255.0,((float)c.green())/255.0,((float)c.blue())/255.0);
		glVertex3f(s->getx(),s->gety(),s->getz());
		glEnd();
	   }
	 else
	   {
		// arete:
//		pos2= * (*table)[(long) b->getAlpha(0)->getSommet()];
//		pos2= (pos1+pos2)/2;
//		p->drawLine(pos1.x(),pos1.y(),pos2.x(),pos2.y());
		sommet *s2 = b->getAlpha(0)->getSommet();
		float 	x = (s->getx()+s2->getx())/2,
		  y = (s->gety()+s2->gety())/2,
		  z = (s->getz()+s2->getz())/2;
		glBegin(GL_LINES);
		glColor3f(((float)c.red())/255.0,((float)c.green())/255.0,((float)c.blue())/255.0);			
		glVertex3f(s->getx(),s->gety(),s->getz());
		glVertex3f(x,y,z);
		glEnd();
	   }
	 
	 b= (brin *) e->suivant(b);
    }
  glLineWidth(1);
}
#undef R
//.........................................
void fenetreGL::dessineGraphe()
{
//  QIntDict<QPoint> * table= plongement();
  
  // Affichage des brins:
  ensemble  *e,  pt;
  QColor * c;
  
  e= (ensemble*) G->classesBrins()->dernier();
  c= couleurs->last();
  
  pt = G->orbite(G->getBrinPointe(),false,false,true,true);
  
  glNewList(1,GL_COMPILE);
  while (e!=NULL)
    {
	 dessineEnsemble(e,&pt,*c);//,table);
	 
	 e= (ensemble *) G->classesBrins()->precedant(e);
	 c= couleurs->prev();
    }
  
  if (G->faceFermee1()) {
    glBegin(GL_LINES);
    glColor3f(C_NORMALE1.red()/255.,C_NORMALE1.green()/255.,C_NORMALE1.blue()/255.);
    glVertex3f(G->barycentreFace1().getx(),G->barycentreFace1().gety(),G->barycentreFace1().getz());
    sommet s = G->barycentreFace1()+G->normaleFace1();
    glVertex3f(s.getx(),s.gety(),s.getz());
    glEnd(); }
  if (G->faceFermee2()) {
    glBegin(GL_LINES);
    glColor3f(C_NORMALE2.red()/255.,C_NORMALE2.green()/255.,C_NORMALE2.blue()/255.);
    glVertex3f(G->barycentreFace2().getx(),G->barycentreFace2().gety(),G->barycentreFace2().getz());
    sommet s = G->barycentreFace2()+G->normaleFace2();
    glVertex3f(s.getx(),s.gety(),s.getz());
    glEnd(); }
  
  // Affichage des sommets:
//  QPoint pos;
  sommet *s= G->sommet1();
//  p->setPen(C_SOMMET);
  while (s!=NULL)
    {
//	 pos= * (*table)[(long) s];
	 
//	 p->drawLine(pos.x()-1,pos.y()  ,pos.x()+1,pos.y()  );
//	 p->drawLine(pos.x()  ,pos.y()-1,pos.x()  ,pos.y()+1);
	 glBegin(GL_POINTS);
	 glColor3f(1.0,1.0,1.0);
	 glVertex3f(s->getx(),s->gety(),s->getz());
	 glEnd();
	 
	 s= G->sommetSuivant(s);
    }
  glEndList();
}
#undef LIGNE

sommet *fenetreGL::normaleFace(brin* b)
{
	sommet *s = new sommet;
	sommet 	*s1 = b->getSommet(),
			*s2 = b->getAlpha(0)->getAlpha(1)->getSommet(),
			*s3 = b->getAlpha(0)->getAlpha(1)->getAlpha(0)->getAlpha(1)->getSommet(),
			*v1 = &((*s1)-(*s2)),
			*v2 = &((*s3)-(*s2));
	s->setx(v1->gety()*v2->getz()-v2->gety()*v1->getz());
	s->sety(v2->getx()*v1->getz()-v1->getx()*v2->getz());
	s->setz(v1->getx()*v2->gety()-v2->getx()*v1->gety());
	delete v1;delete v2;
	return s;
}

sommet *fenetreGL::normaleVertex(brin *b)
{
	sommet *s = new sommet(0,0,0);
	brin *b1 = b;
	int i=0;
	do {
		i++;
		sommet *ss = normaleFace(b1),*ss2;
		ss2 = &(*s + *ss);
		*s = *ss2;
		delete ss;delete ss2;
		b1 = b1->getAlpha(1)->getAlpha(2);
	} while (b1!=b);
//	*s /= i;
	return s;
}


void fenetreGL::dessinePlein()
{
  if (G->brin1()==NULL) {	glNewList(3,GL_COMPILE);glEndList();return; }
  
  glNewList(3,GL_COMPILE);
  if (Blend) { 	glEnable(GL_BLEND);
  glDisable(GL_DEPTH_TEST); }
//	glColor4f(0,0,1,0.3);

  ensemble total;
  total = (G->ensembleBrins());
  while (!total.estVide()) {
    ensemble f = G->orbiteFace((brin*)(total.choisis()));
    total.enleveEnsemble(f);
    // process face f
    if (f.taille() < 6) continue;
    // normales
    brin *b = (brin*)(f.premier());
/*    GLdouble v1x=b->getSommet()->getx(),v1y=b->getSommet()->gety(),v1z=b->getSommet()->getz();
    b = b->getAlpha(0);b = b->getAlpha(1);
    GLdouble v2x=b->getSommet()->getx(),v2y=b->getSommet()->gety(),v2z=b->getSommet()->getz();
    b = b->getAlpha(0);b = b->getAlpha(1);
    GLdouble v3x=b->getSommet()->getx(),v3y=b->getSommet()->gety(),v3z=b->getSommet()->getz();
    GLdouble 	vv1x=v2x-v1x,vv1y=v2y-v1y,vv1z=v2z-v1z,
	 vv2x=v3x-v1x,vv2y=v3y-v1y,vv2z=v3z-v1z;
    GLdouble n[3];
    n[0]=(vv1y*vv2z)-(vv2y*vv1z);
    n[1]=(vv2x*vv1z)-(vv1x*vv2z);
    n[2]=(vv1x*vv2y)-(vv2x*vv1y);

    // eye coords
    GLdouble ec[3] = { sin(3.14*alpha/180), cos(3.14*alpha/180), sin(3.14*beta/180) };
    if (ec[0]*n[0]+ec[1]*n[1]+ec[2]*n[2]>0) {n[0]=-n[0];n[1]=-n[1];n[2]=-n[2];}
    //

//		printf("%f,%f,%f\n",n[0],n[1],n[2]);
    
    b = (brin*)(f.premier());
*/
    GLdouble *face = new GLdouble[f.taille()/2*6];
    for (int i=0;i<f.taille()/2;i++) {
	 face[6*i] = (GLdouble)(b->getSommet()->getx());
	 face[6*i+1] = (GLdouble)(b->getSommet()->gety());
	 face[6*i+2] = (GLdouble)(b->getSommet()->getz());
	
	 sommet *s = normaleVertex(b);
	 face[6*i+3] = (GLdouble)(s->getx());
	 face[6*i+4] = (GLdouble)(s->gety());
	 face[6*i+5] = (GLdouble)(s->getz());
	 delete s;
	 //b = (brin*)(f->suivant(b));
	 b = b->getAlpha(0);
	 b = b->getAlpha(1);
//			printf("%f,%f,%f\n",face[3*i],face[3*i+1],face[3*i+2]);
    }
    // display face
    if (b->getSommet()==((brin*)(f.premier()))->getSommet())
	    tShade((int)(f.taille())/2,face,(Lighting)?1:0);
    delete face;
  }
  
  if (Blend) {	glDisable(GL_BLEND);
  glEnable(GL_DEPTH_TEST); }
  glEndList();
}

#define K1 (0.9)
#define K2 (0.8)
void fenetreGL::dessineEclate()
{
  if (G->brin1()==NULL) {	glNewList(4,GL_COMPILE);glEndList();return; }
  
  glNewList(4,GL_COMPILE);
  
  glColor4f(1,1,1,1);
  
  ensemble total;
  total = (G->ensembleBrins());
  while (!total.estVide()) {
    ensemble f = G->orbiteFace((brin*)(total.choisis()));
    total.enleveEnsemble(f);
    // process face f
    // calcul du barycentre			
    glBegin(GL_LINES);
    glColor4f(1,1,1,1);
    brin *b = (brin*)(f.premier());
    coord x=0,y=0,z=0;
    for (int i=0;i<f.taille()/2;i++) {
	 x+=b->getSommet()->getx();
	 y+=b->getSommet()->gety();
	 z+=b->getSommet()->getz();
	
	 if (b->getAlpha(1)!=NULL) {
	 	sommet 	*s1 = b->getAlpha(0)->getSommet(),
	 			*s3 = b->getAlpha(0)->getAlpha(1)->getAlpha(0)->getSommet(),
	 			s = *s1 + 0.2*(*(b->getSommet()) - *s1),
	 			ss = *s1 + 0.2*(*s3 - *s1);
	 	glVertex3f(s.getx(),s.gety(),s.getz());
	 	glVertex3f(ss.getx(),ss.gety(),ss.getz());
	 }
	
	 b = b->getAlpha(0);
	 b = b->getAlpha(1);
    }
    glEnd();
    x/=f.taille()/2;
    y/=f.taille()/2;
    z/=f.taille()/2;				
    if (b->getSommet()!=((brin*)(f.premier()))->getSommet()) continue;
    glBegin(GL_LINES);		
    b = (brin*)(f.premier());
    for (int i=0;i<f.taille()/2;i++) {
	 glColor4f(0.95,0.95,0.95,1.0);
	 glVertex3f(	x+ (b->getSommet()->getx()-x)*K1,
				y+ (b->getSommet()->gety()-y)*K1,
				z+ (b->getSommet()->getz()-z)*K1 );
	 brin *nb11,*nb1 = b->getAlpha(2);if (nb1==b) nb1=NULL; else nb11 = b;
	 b = b->getAlpha(0);
	 glVertex3f(	x+ (b->getSommet()->getx()-x)*K1,
				y+ (b->getSommet()->gety()-y)*K1,
				z+ (b->getSommet()->getz()-z)*K1 );
	 brin *nb22,*nb2 = b->getAlpha(2);if (nb2==b) nb2=NULL; else nb22 = b;
	 b = b->getAlpha(1);
	 glColor4f(0.9,0.9,0.9,1.0);			
	 if (nb1!=NULL) {
	   glVertex3f(	(x+ (nb11->getSommet()->getx()-x)*K1) + 0.25*(  -(x+ (nb11->getSommet()->getx()-x)*K1) + (x+ (nb11->getAlpha(0)->getSommet()->getx()-x)*K1) ),
				(y+ (nb11->getSommet()->gety()-y)*K1) + 0.25*(  -(y+ (nb11->getSommet()->gety()-y)*K1) + (y+ (nb11->getAlpha(0)->getSommet()->gety()-y)*K1) ),
				(z+ (nb11->getSommet()->getz()-z)*K1) + 0.25*(  -(z+ (nb11->getSommet()->getz()-z)*K1) + (z+ (nb11->getAlpha(0)->getSommet()->getz()-z)*K1)  ) );
	   glVertex3f(	nb1->getSommet()->getx() + 0.25*(nb1->getAlpha(0)->getSommet()->getx() - nb1->getSommet()->getx()),
				nb1->getSommet()->gety() + 0.25*(nb1->getAlpha(0)->getSommet()->gety() - nb1->getSommet()->gety()),
				nb1->getSommet()->getz() + 0.25*(nb1->getAlpha(0)->getSommet()->getz() - nb1->getSommet()->getz()) );
	 }
	 if (nb2!=NULL) {
	   glVertex3f(	(x+ (nb22->getSommet()->getx()-x)*K1) + 0.25*(  -(x+ (nb22->getSommet()->getx()-x)*K1) + (x+ (nb22->getAlpha(0)->getSommet()->getx()-x)*K1) ),
				(y+ (nb22->getSommet()->gety()-y)*K1) + 0.25*(  -(y+ (nb22->getSommet()->gety()-y)*K1) + (y+ (nb22->getAlpha(0)->getSommet()->gety()-y)*K1) ),
				(z+ (nb22->getSommet()->getz()-z)*K1) + 0.25*(  -(z+ (nb22->getSommet()->getz()-z)*K1) + (z+ (nb22->getAlpha(0)->getSommet()->getz()-z)*K1) ) );
	   glVertex3f(	nb2->getSommet()->getx() + 0.25*(nb2->getAlpha(0)->getSommet()->getx() - nb2->getSommet()->getx()),
				nb2->getSommet()->gety() + 0.25*(nb2->getAlpha(0)->getSommet()->gety() - nb2->getSommet()->gety()),
				nb2->getSommet()->getz() + 0.25*(nb2->getAlpha(0)->getSommet()->getz() - nb2->getSommet()->getz()) );
	 }
	 
	 
    }
    glEnd();
  }
  glEndList();
}


//.........................

void fenetreGL::majBuffer()
{
  dessineGraphe();
  if (Shade) dessinePlein();
  if (Eclate) dessineEclate();
  repaint(false);
  emit raffraichissement(this);
}

//======================================================================================================

void fenetreGL::initializeGL()
{
  // OpengGL initialization
  
  ViewportHeight = height();//400;
  ViewportWidth = width();//450;
  
  glEnable(GL_POLYGON_MODE);	
  glPolygonMode(GL_FRONT,GL_FILL);
  glPolygonMode(GL_BACK,GL_FILL);
//	glCullFace(GL_BACK);
//	glEnable(GL_CULL_FACE);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
  glBlendEquation(GL_FUNC_ADD);
  
  
  glEnable(GL_DEPTH_TEST);
  glClearColor(0.0, 0.0, 0.0, 1.0);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-ViewWidth/2, ViewWidth/2,
		  -(ViewWidth/2)*(((float)ViewportHeight)/((float)ViewportWidth)),
		  (ViewWidth/2)*(((float)ViewportHeight)/((float)ViewportWidth)), 1.0, 50.0);
  
//	glDrawBuffer(GL_BACK);
//	glReadBuffer(GL_BACK);
  
  glPointSize(4);
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glRotatef(-90,1,0,0);
  glRotatef(alpha,0,0,1);
  glRotatef(beta,-cos(alpha*3.14/180),sin(alpha*3.14/180),0);		
  glTranslatef(rayon*sin(alpha*3.14/180)*cos(beta*3.14/180)-xV,
			rayon*cos(alpha*3.14/180)*cos(beta*3.14/180)-yV,
			rayon*sin(beta*3.14/180)-zV);
  glScalef(kmul,kmul,kmul);	
  
  
  glHint(GL_FOG_HINT,GL_NICEST);
  
  GLfloat fogcolor[4] = { 0.,0.,0.,0. };
  
  glFogi(GL_FOG_MODE,GL_EXP2);
  glFogfv(GL_FOG_COLOR,fogcolor);
  glFogf(GL_FOG_DENSITY,0.050f);
  glFogf(GL_FOG_START,48.f);
  glFogf(GL_FOG_END,50.f);
  glEnable(GL_FOG);
  
  dessineQuadrillage();
}

void fenetreGL::paintGL()
{
  if (Antialias) {
    glClear(GL_ACCUM_BUFFER_BIT);
    for (int jitter = 0; jitter < 8; jitter++)
	 {
	   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	   
	   accFrustum(-ViewWidth/2, ViewWidth/2,
			    -(ViewWidth/2)*(((float)ViewportHeight)/((float)ViewportWidth)),
			    (ViewWidth/2)*(((float)ViewportHeight)/((float)ViewportWidth)), 1.0, 50.0,
			    j8[jitter].x, j8[jitter].y, 0.0, 0.0, 1.0);
	   
	   //////
	   glCallList(2);		// quadrillage
	   if (Shade)
		glCallList(3);		// shading
	   if (Eclate)
		glCallList(4);		// vue eclate
	   if (!Shade||Blend)
		glCallList(1);		// modle	1
	   //////
	   glAccum(GL_ACCUM, 1.0/8.);
	 }
    glAccum (GL_RETURN, 1.0);
    Antialias = false;
  } else {
    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
    glCallList(2);		// quadrillage
    if (Shade)
	 glCallList(3);		// shading
    if (Eclate)
	 glCallList(4);		// vue eclate
    if (!Shade||Blend)
	 glCallList(1);		// modle	1
  }
  
}

void fenetreGL::resizeGL(int w,int h)
{
  ViewportWidth = w;
  ViewportHeight = h;
  glViewport(0, 0, w, h);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glFrustum(-ViewWidth/2, ViewWidth/2,
		  -(ViewWidth/2)*(((float)ViewportHeight)/((float)ViewportWidth)),
		  (ViewWidth/2)*(((float)ViewportHeight)/((float)ViewportWidth)), 1.0, 50.0);
  
  repaint(false);
}


float fenetreGL::getAlpha() {return alpha;};
void fenetreGL::setAlpha(float a)
{
  
  a = a-((int)(a/360))*360;
  if (a<0) a+=360;
  if (a!=alpha)
    {
	 alpha=a;
	 
	 glMatrixMode(GL_MODELVIEW);
	 glLoadIdentity();
	 glRotatef(-90,1,0,0);
	 glRotatef(alpha,0,0,1);
	 glRotatef(beta,-cos(alpha*3.14/180),sin(alpha*3.14/180),0);		
	 glTranslatef(rayon*sin(alpha*3.14/180)*cos(beta*3.14/180)-xV,
			    rayon*cos(alpha*3.14/180)*cos(beta*3.14/180)-yV,
			    rayon*sin(beta*3.14/180)-zV);
	 glScalef(kmul,kmul,kmul);	
	 maj();
    }
}
float fenetreGL::getBeta () {return beta;};
void fenetreGL::setBeta (float b)
{
  b = b-((int)(b/360))*360;
  if (b<0) b+=360;
  if ((b<270)&&(b>90))
    if (b<180) b=90; else b=270;
  if (b!=beta)
    {
	 beta=b;
	 glMatrixMode(GL_MODELVIEW);
	 glLoadIdentity();
	 glRotatef(-90,1,0,0);
	 glRotatef(alpha,0,0,1);
	 glRotatef(beta,-cos(alpha*3.14/180),sin(alpha*3.14/180),0);		
	 glTranslatef(rayon*sin(alpha*3.14/180)*cos(beta*3.14/180)-xV,
			    rayon*cos(alpha*3.14/180)*cos(beta*3.14/180)-yV,
			    rayon*sin(beta*3.14/180)-zV);
	 glScalef(kmul,kmul,kmul);
	 maj();
    }
}

void fenetreGL::setAlphaBeta(float a,float b)
{
  a = a-((int)(a/360))*360;
  if (a<0) a+=360;
  
  b = b-((int)(b/360))*360;
  if (b<0) b+=360;
  if ((b<270)&&(b>90))
    if (b<180) b=90; else b=270;
  
  if ((a!=alpha)||(b!=beta))
    {
	 alpha=a;
	 beta=b;
	 glMatrixMode(GL_MODELVIEW);
	 glLoadIdentity();
	 glRotatef(-90,1,0,0);
	 glRotatef(alpha,0,0,1);
	 glRotatef(beta,-cos(alpha*3.14/180),sin(alpha*3.14/180),0);
	 glTranslatef(rayon*sin(alpha*3.14/180)*cos(beta*3.14/180)-xV,
			    rayon*cos(alpha*3.14/180)*cos(beta*3.14/180)-yV,
			    rayon*sin(beta*3.14/180)-zV);
	 glScalef(kmul,kmul,kmul);
	 maj();
    }
}

float fenetreGL::getK    () {return kmul;}
void fenetreGL::setK    (float k)
{
  if (k>10.0) k=10.0;
  if (k<0.1) k=0.1;
  if (k!=kmul)
    {
	 kmul=k;
	 glMatrixMode(GL_MODELVIEW);
	 glLoadIdentity();
	 glRotatef(-90,1,0,0);
	 glRotatef(alpha,0,0,1);
	 glRotatef(beta,-cos(alpha*3.14/180),sin(alpha*3.14/180),0);		
	 glTranslatef(rayon*sin(alpha*3.14/180)*cos(beta*3.14/180)-xV,
			    rayon*cos(alpha*3.14/180)*cos(beta*3.14/180)-yV,
			    rayon*sin(beta*3.14/180)-zV);
	 glScalef(kmul,kmul,kmul);
	 maj();
    }
}

void fenetreGL::getPos(float *px,float *py,float *pz)
{
  *px = xV;
  *py = yV;
  *pz = zV;
}

void fenetreGL::setPos(float x,float y,float z)
{
  xV = x;
  yV = y;
  zV = z;
}

void fenetreGL::setShade(bool b) { Shade = b; majBuffer();}
bool fenetreGL::getShade() { return Shade; }
void fenetreGL::setBlend(bool b) { Blend = b; majBuffer();}
bool fenetreGL::getBlend() { return Blend; }
void fenetreGL::setEclate(bool b) { Eclate = b; majBuffer();}
bool fenetreGL::getEclate() { return Eclate; }
void fenetreGL::setLighting(bool b) { Lighting = b; majBuffer();}
bool fenetreGL::getLighting() { return Lighting; }

void fenetreGL::DisplayWithAntialias(bool b)
{
  Antialias = b;
  repaint(false);
}
