#include <math.h> // pour les fonctions trigonomtriques (dans les rotations)
#include <qfile.h>
#include <qdict.h>
#include <qobject.h>
#include <qregexp.h>

#include "base.h"
#include "gcarte.h"

gcarte::gcarte()
{
  modif   = false;
}

gcarte::~gcarte()
{
}

void gcarte::slotBModifie(brin * b) // attention: on peut avoir b==NULL
{
  modif=true;
  emit carteModifiee(this);

  if (b!=NULL)
    {
	 sommet * s = b->getSommet();
	 // Si le sommet sur lequel pointe le brin  t modifi,
	 // il se peut que celui-ci ne soit pas encore dans la gcarte! => on le rajoute:
	 if (!ES.contient(s))
	   {
		ES.ajoute(s);
		connect(s,SIGNAL(modifie(sommet *)),this,SLOT(slotSModifie(sommet *)));
		connect(s,SIGNAL(isole  (sommet *)),this,SLOT(slotIsole   (sommet *)));
	   }
    }
}

void gcarte::slotIsole(sommet * s)
{
  // Le sommet s est isol (aucun brin ne pointe sur lui)
  // On le supprime de la liste: (la destruction est automatique)
  ES.enleve(s);
}
 
void gcarte::slotSModifie(sommet * s) // attention: on peut avoir S==NULL
{
  modif=true;
  emit carteModifiee(this);
}

brin * gcarte::ajouteBrin(sommet * s)
{
  brin * b= new brin(s);
  ES.ajoute(s);
  EB.ajoute(b);

  connect(s,SIGNAL(modifie(sommet *)),this,SLOT(slotSModifie(sommet *)));
  connect(s,SIGNAL(isole  (sommet *)),this,SLOT(slotIsole   (sommet *)));
  connect(b,SIGNAL(modifie(brin   *)),this,SLOT(slotBModifie(brin   *)));

  slotBModifie(b);
  return b;
}

void gcarte::supprimeBrin(brin * b)
{
  if (!EB.contient(b))
    erreur("gcarte[supprimeBrin]: Vous essayer d'enlever un brin qui n'appartient pas  l'ensemble.");

  EB.enleve(b);
  delete b;
}

brin * gcarte::dupliqueComposante(brin * bsel) // brin slectionn
{
  if (bsel==NULL)
    return NULL;

  QIntDict<sommet> S;
  QIntDict<brin  > B;

  ensemble e= orbiteComposante(bsel);
  brin * b;
  // Nouveaux brins et sommets:
  b= (brin *) e.premier();
  while (b!=NULL)
    {
	 sommet * s= b->getSommet();

	 if (S[(unsigned long) s]==NULL)
	   S.insert((unsigned long) s, new sommet(*s));

	 B.insert((unsigned long) b, ajouteBrin(S[(unsigned long) s]));

	 b= (brin *) e.suivant(b);
    }
  // Coutures:
  brin * bb; int i;
  b= (brin *) e.premier();
  while (b!=NULL)
    {
	 bb= B[(unsigned long) b];

	 for (i=0; i<=3; i++)
	   bb->setAlpha(i,B[(unsigned long) b->getAlpha(i)]);

	 b= (brin *) e.suivant(b);
    }

  //  bsel= B[(unsigned long) bsel];
  return bsel;
}

void gcarte::supprimeComposante(brin * b)
{
  ensemble e= orbiteComposante(b);

  while (!e.estVide())
    supprimeBrin((brin *) e.choisis());
}

brin * gcarte::brin1() const
{
  return (brin *) EB.premier();
}

sommet * gcarte::sommet1() const
{
  return (sommet *) ES.premier();
}

brin * gcarte::brinSuivant(const brin * b) const
{
  return (brin *) EB.suivant(b);
}

sommet * gcarte::sommetSuivant(const sommet * s) const
{
  return (sommet *) ES.suivant(s);
}

unsigned int gcarte::nbBrins  () const
{
  return EB.taille();
}

unsigned int gcarte::nbSommets() const
{
  return ES.taille();
}

bool gcarte::modifiee() const
{
  return modif;
}

void gcarte::nouvelle()
{
  if (nbSommets()!=0)
    {
	 EB.vide(true);
	 modif=false;
	 emit carteModifiee(this);
    }
}

// -------------------------------------------------------------------------------
// Les orbites:
ensemble gcarte::orbite(const brin *b, bool a0, bool a1, bool a2, bool a3) const
{
  ensemble explores, aEtudier;

  if (b==NULL)
    return explores;
  
  brin *v; // v comme voisin
  
  aEtudier.ajoute(b);

  while (!aEtudier.estVide())
    {
	 b= (brin *) aEtudier.choisis();
	 explores.ajoute(b);

	 if (a0) { v= b->getAlpha(0); if (!explores.contient(v)) aEtudier.ajoute(v); }
	 if (a1) { v= b->getAlpha(1); if (!explores.contient(v)) aEtudier.ajoute(v); }
	 if (a2) { v= b->getAlpha(2); if (!explores.contient(v)) aEtudier.ajoute(v); }
	 if (a3) { v= b->getAlpha(3); if (!explores.contient(v)) aEtudier.ajoute(v); }
    }

  return explores;
}

ensemble gcarte::orbiteSommet(const brin * b) const
{
  return orbite(b,false,true ,true ,true );
}

ensemble gcarte::orbiteFace(const brin * b) const
{
  return orbite(b,true ,true ,false,false);
}

ensemble gcarte::orbiteVolume(const brin * b) const
{
  return orbite(b,true ,true ,true ,false);
}

ensemble gcarte::orbiteComposante(const brin * b) const
{
  return orbite(b,true ,true ,true ,true );
}

ensemble gcarte::orbiteArete(const brin * b) const
{
  return orbiteSommet(b).ajouteEnsemble(orbiteSommet(b->getAlpha(0)));
}

ensemble gcarte::ensembleBrins() const
{
  return EB;
}
//-----------------------------------------------------------------------------------
// On suppose la face ferme!
coord gcarte::circonferenceFace(const brin * b0) const
{
  const brin * b = b0;
  const brin * c = b0->getAlpha(0)->getAlpha(1);

  coord D=0;

  do
    {
	 D+= (*b->getSommet() - *c->getSommet()).norme();

	 b= c;
	 c= c->getAlpha(0)->getAlpha(1);
    }
  while (b!=b0);

  return D;
}
//-----------------------------------------------------------------------------------
void gcarte::ajouteSommet(brin * b, sommet s)
{
  sommet   * ns = new sommet(s);
  ensemble   e  = orbite(b, false,false,true,true);

  unsigned int i,n= e.taille();
  brin * T[n][2];


  // Cration des brins:
  for (i=0; i<n; i++)
    {
	 T[i][0]= ajouteBrin(ns);
	 T[i][1]= ajouteBrin(ns);
    }
  // Coutures par alpha0 et alpha1:
  b= (brin *) e.premier();
  for (i=0; i<n; i++)
    {
	 brin * c= b->getAlpha(0);
	 
	 T[i][0]->setAlpha(0,b); b->setAlpha(0,T[i][0]);
	 T[i][1]->setAlpha(0,c); c->setAlpha(0,T[i][1]);

	 T[i][0]->setAlpha(1,T[i][1]);
	 T[i][1]->setAlpha(1,T[i][0]);

	 b= (brin *) e.suivant(b);
    }
  // Coutures par alpha2 et alpha3:
  b= (brin *) e.premier();
  for (i=0; i<n; i++)
    {
	 brin * c= b->getAlpha(0)->getAlpha(1)->getAlpha(0);
	 
	 T[i][0]->setAlpha(2,b->getAlpha(2)->getAlpha(0));
	 T[i][1]->setAlpha(2,c->getAlpha(2)->getAlpha(0));
	 T[i][0]->setAlpha(3,b->getAlpha(3)->getAlpha(0));
	 T[i][1]->setAlpha(3,c->getAlpha(3)->getAlpha(0));

	 b= (brin *) e.suivant(b);
    }
}
//-----------------------------------------------------------------------------------
bool gcarte::adjonctionAretePossible(const brin * b1, const brin * b2)
{
  if (b1==NULL || b2==NULL)
    return false;

  ensemble E= orbiteFace(b1);

  E.enleveEnsemble(orbiteSommet(b1                          ));
  E.enleveEnsemble(orbiteSommet(b1             ->getAlpha(0)));
  E.enleveEnsemble(orbiteSommet(b1->getAlpha(1)->getAlpha(0)));

  return E.contient(b2);
}

void gcarte::ajouteArete(brin * b1, brin * b2)
{
  if (!adjonctionAretePossible(b1,b2))
    return;

  // Faire en sorte que b1 et b2 se retrouvent sur la mme face:
  brin * b= b1;
  do
    b= b->getAlpha(0)->getAlpha(1);
  while (b!=b1 && b!=b2);

  if (b==b2)
    b2= b2->getAlpha(1);

  // Adjonction:
  brin * b11= b1->getAlpha(1);
  brin * b21= b2->getAlpha(1);
  brin * B[2][2];

  B[0][0]= ajouteBrin(b1->getSommet());
  B[0][1]= ajouteBrin(b1->getSommet());
  B[1][0]= ajouteBrin(b2->getSommet());
  B[1][1]= ajouteBrin(b2->getSommet());

  B[0][0]->setAlpha(0,B[1][0]);
  B[1][0]->setAlpha(0,B[0][0]);
  B[0][1]->setAlpha(0,B[1][1]);
  B[1][1]->setAlpha(0,B[0][1]);

  B[0][0]->setAlpha(2,B[0][1]);
  B[0][1]->setAlpha(2,B[0][0]);
  B[1][0]->setAlpha(2,B[1][1]);
  B[1][1]->setAlpha(2,B[1][0]);

  b1     ->setAlpha(1,B[0][0]);
  B[0][0]->setAlpha(1,b1     );
  b11    ->setAlpha(1,B[0][1]);
  B[0][1]->setAlpha(1,b11    );
  b2     ->setAlpha(1,B[1][0]);
  B[1][0]->setAlpha(1,b2     );
  b21    ->setAlpha(1,B[1][1]);
  B[1][1]->setAlpha(1,b21    );
  
}
//-----------------------------------------------------------------------------------
ensemble gcarte::brinsProches(const brin * b0) const
{
  sommet s0= * b0->getSommet();
  sommet s1= * b0->getAlpha(0)->getSommet();

  ensemble E;
  brin * b= brin1();
  while (b!=NULL)
    {
	 if ((s0-*b->             getSommet()).norme()<epsilon &&
		(s1-*b->getAlpha(0)->getSommet()).norme()<epsilon)
	   E.ajoute(b);

	 b= brinSuivant(b);
    }

  return E;
}
//-----------------------------------------------------------------------------------
bool gcarte::orientable(const brin * b) const
{
  ensemble explores, aEtudier, composante= orbiteComposante(b);

  if (b==NULL)
    return false;
  
  brin *v; // v comme voisin
  
  aEtudier.ajoute(b);

  while (!aEtudier.estVide())
    {
	 b= (brin *) aEtudier.choisis();
	 explores.ajoute(b);

	 for (int i=1; i<=3; i++)
	   if (b->getAlpha(0)->getAlpha(i)!=b->getAlpha(0))
		{
		  v= b->getAlpha(0)->getAlpha(i);
		  if (!explores.contient(v))
		    aEtudier.ajoute(v);
		}
    }

  return explores.taille() < composante.taille();
}
//-----------------------------------------------------------------------------------
unsigned int gcarte::genre(const brin * b) const
{
  if (orientable(b))
    return 0;
  else
    {
	 ensemble e = orbiteComposante(b);

	 if ((e.taille()+sommets(e).taille()) % 2 == 1)
	   return 1;
	 else
	   return 2;
    }
}
//-----------------------------------------------------------------------------------
#include "sauver-charger.cc"
#include "coutures.cc"
#include "geometrie.cc"
#include "decoutures.cc"
#include "objets.cc"
#include "extrusion.cc"
#include "extrusion_multiple.cc"
#include "couture_pb.cc"
#include "fusion.cc"
#include "ajouteFace.cc"
//-----------------------------------------------------------------------------------
