#ifndef _gcarte_h
#define _gcarte_h

#include <qobject.h>
#include <qlist.h>
#include <qfile.h>

#include "ensemble.h"
#include "sommet.h"
#include "brin.h"

typedef enum
{
  Aucun, /* Cette valeur ne doit pas etre modifie (pour la validit des tests boolens) */
  Sommets, Aretes, Faces, Volumes
} objets;

class gcarte: public QObject
{
  Q_OBJECT
    
    private:
  // Ces ensembles sont des ensembles de pointeurs vers des sommets/des brins.
  ensemble ES;
  ensemble EB;
  bool modif;
  
  // Mthodes auxiliaires pour la mthode charger:
  bool decimal(QString);
  bool id(QString);
  QString token(QFile *);
  
  // Mthodes pour coutures gomtriques:
  void orienteObjet      (const brin *, const brin *);
  void redimensionneObjet(const brin *, const brin *);
  void deplaceObjet      (const brin *, const brin *);
  // Pour l'extrusion de face: Vrifie  quoi appartient le brin
  // (sommet seul, arete, face seule, ou volume.
  int app_safv(brin *);
  
 signals:
  void carteModifiee(gcarte *);
  
  public slots:
    // slot pour le signal reu quand un sommet se retrouve isol (aucun brin ne pointe sur lui).
    void slotIsole(sommet *);
    // slots pour le signal reu quand un sommet ou un brin est modifi.
    void slotSModifie(sommet *);
    void slotBModifie(brin   *);
  
 public:
  gcarte();
  ~gcarte();
  
  brin * ajouteBrin(sommet *);
  void   supprimeBrin(brin *);
  
  brin * dupliqueComposante(brin *);
  void   supprimeComposante(brin *);
  
  // retourne le 1er brin   de la carte, NULL si pas de brin  .
  brin   * brin1  () const;
  // retourne le 1er sommet de la carte, NULL si pas de sommet.
  sommet * sommet1() const;
  // retourne le brin   suivant celui pass en paramtre, NULL si pas de suivant.
  brin   * brinSuivant  (const brin   *) const;
  // retourne le sommet suivant celui pass en paramtre, NULL si pas de suivant.
  sommet * sommetSuivant(const sommet *) const;
  
  // Nombre de brins et de sommets dans le graphe:
  unsigned int nbBrins  () const;
  unsigned int nbSommets() const;
  
  bool modifiee() const;
  void nouvelle();
  
  // Mthodes de couture et de dcouture:
  objets couturePossible  (const brin *, const brin *) const;
  objets decouturePossible(const brin *) const;
  
  bool   coudre  (brin *, brin *, bool =false, bool =false, bool =false);
  brin * decoudre(brin *);
  
  // Mthodes de chargement et de sauvegarde:
  bool         sauvegarder(const QString);
  unsigned int charger    (const QString /* nomFic */,QString & /* message d'erreur */, bool raz /* remise  zro */=true);
  
  // Mthode permettant de savoir si un brin donn est sur une face FERMEE:
  bool faceFermee(const brin *) const;
  // Mthode permettant de trouver la normale d'une face:
  // Attention: Si la face n'est pas ferme, renvoie le vecteur (0,0,1)
  sommet normale(const brin *) const;
  
  // Parcours:
  ensemble orbite          (const brin *, bool,bool,bool,bool) const;
  
  ensemble orbiteSommet    (const brin *) const;
  ensemble orbiteFace      (const brin *) const;
  ensemble orbiteVolume    (const brin *) const;
  ensemble orbiteComposante(const brin *) const;  
  ensemble orbiteArete     (const brin *) const;
  ensemble ensembleBrins() const;
  
  // Dcoupage d'une arte en deux:
  void ajouteSommet(brin *, const sommet);
  // Adjonction d'une nouvelle arte entre deux sommets:
  bool adjonctionAretePossible(const brin *, const brin *);
  void ajouteArete (brin *, brin *);
  brin * ajouteFace(ensemble);
  
  // Circonfrence d'une face:
  coord circonferenceFace(const brin *) const;
  
  // Cration d'objets de base:
  brin * insereFace(int);
  brin * inserePyramide(int); // nombre d'aretes de la base
  brin * insereCube();
  brin * insereSphere(int,int); // nombre mridiens et de parallles
  brin * insereTore(coord,coord,int,int); // R, r, mridiens et parallles

  // Test d'orientabilit d'un objet:
  bool         orientable(const brin *) const;
  unsigned int genre     (const brin *) const;

  // Extrusions:
  brin * extrusionFace            (      brin * /* face */, const sommet   /* vecteur */, const unsigned int /* nombre de mridiens */ =3);
  bool   extrusionMultiplePossible(const brin * /* face */, const brin   * /* chemin  */) const;
  brin * extrusionMultiple        (      brin * /* face */, const brin   * /* chemin  */);
  // Fusion de Volumes:
  brin * fusion(brin *);
  // Couture par Bords:
  brin * coutureParBord(brin *, brin *);

  // Pour la slection des superposs:
  ensemble brinsProches(const brin *) const;
};



// Fonctions utiles:
//------------------

// Ensemble de sommets correspondant  un ensemble de brins:
ensemble & sommets(const ensemble &);
// Barycentre d'un ensemble de sommets:
sommet barycentre(const ensemble &);
// Dplacement, Agrandissement, Rotation (angles en RADIANS) d'ensembles de SOMMETS:
void   translation   (sommet,                         const ensemble &, const ensemble &);
void   agrandissement(sommet, sommet,                 const ensemble &, const ensemble &);
void   rotation      (sommet, sommet,                 const ensemble &, const ensemble &);
void   orientation   (sommet, sommet, sommet, sommet, const ensemble &, const ensemble &);

sommet angles        (sommet, sommet);

void   translation   (sommet,                         const ensemble &);
void   agrandissement(sommet, sommet,                 const ensemble &);
void   rotation      (sommet, sommet,                 const ensemble &);
void   orientation   (sommet, sommet, sommet, sommet, const ensemble &);

sommet rotation(const sommet, const sommet);
coord  /* radians */ angle(const coord /* x */, const coord /* y */);

// Sous-fonction de la fonction extrusionMultiple,
// trouve la base (vecteurs vitesse, normale et acclration) pour un brin donn du chemin
void base(const brin *, sommet &, sommet &, sommet &);

#endif
