//-----------------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
ensemble & sommets(const ensemble & EB)
{
  ensemble * ES = new ensemble;
  brin * b = (brin *) EB.premier();
  
  while (b!=NULL)
    {
	 ES->ajoute(b->getSommet());
	 b= (brin *) EB.suivant(b);
    }
  
  return *ES;
}
//-----------------------------------------------------------------------------------
sommet barycentre(const ensemble & ES)
{
  sommet b(0,0,0);
  sommet * s = (sommet *) ES.premier();
  
  if (s==NULL)
    erreur("gcarte[barycentre]: L'ensemble de sommets est vide!");
  
  while (s!=NULL)
    {
	 b+= *s;
	 s= (sommet *) ES.suivant(s);
    }
  
  return b/ES.taille();
}
//-----------------------------------------------------------------------------------
void translation(sommet vecteur, const ensemble & dep, const ensemble & arr)
{
  sommet * s = (sommet *) dep.premier();
  sommet * t = (sommet *) arr.premier();
  
  while (s!=NULL)
    {
	 *t = *s + vecteur;
	 
	 s= (sommet *) dep.suivant(s);
	 t= (sommet *) arr.suivant(t);
    }
}

void translation(sommet vecteur, const ensemble & dep)
{
  translation(vecteur, dep, dep);
}
//-----------------------------------------------------------------------------------
void agrandissement(sommet coefs, sommet centre, const ensemble & dep, const ensemble & arr)
{
  sommet * s = (sommet *) dep.premier();
  sommet * t = (sommet *) arr.premier();
  
  while (s!=NULL)
    {
	 *t= centre + coefs.produit(*s - centre);
	 
	 s= (sommet *) dep.suivant(s);
	 t= (sommet *) arr.suivant(t);
    }
}

void agrandissement(sommet coefs, sommet centre, const ensemble & dep)
{
  agrandissement(coefs, centre, dep, dep);
}
//-----------------------------------------------------------------------------------
// Les angles sont en RADIANS
void rotation(sommet angles, sommet centre, const ensemble & dep, const ensemble & arr)
{
  sommet * s = (sommet *) dep.premier();
  sommet * t = (sommet *) arr.premier();
  
  coord
    sa = sin(angles.getx()), ca= cos(angles.getx()),
    sb = sin(angles.gety()), cb= cos(angles.gety()),
    sc = sin(angles.getz()), cc= cos(angles.getz()),
    x1,y1,z1,
    x2,y2,z2;
  
  while (s!=NULL)
    {
	 x1= s->getx()-centre.getx();
	 y1= s->gety()-centre.gety();
	 z1= s->getz()-centre.getz();
	 // Rotation sur Ox:
	 x2= x1;
	 y2= y1*ca+z1*sa;
	 z2= z1*ca-y1*sa;
	 // Rotation sur Oy:
	 x1= x2*cb-z2*sb;
	 y1= y2;
	 z1= z2*cb+x2*sb;
	 // Rotation sur Oz:
	 x2= x1*cc+y1*sc;
	 y2= y1*cc-x1*sc;
	 z2= z1;
	 
	 t->setxyz(x2+centre.getx(),
			 y2+centre.gety(),
			 z2+centre.getz());
	 
	 s= (sommet *) dep.suivant(s);
	 t= (sommet *) arr.suivant(t);
    }
}

void rotation(sommet angles, sommet centre, const ensemble & dep)
{
  rotation(angles, centre, dep, dep);
}
//-----------------------------------------------------------------------------
sommet rotation(sommet a /* angles */, sommet v /* vecteur */)
{
  sommet w=v;
  ensemble E,F;
  
  E.ajoute(&v);
  F.ajoute(&w);
  rotation(a,sommet(0,0,0),E,F);
  
  return * (sommet *) F.premier();
}
//-----------------------------------------------------------------------------
// angles: alpha autour de Ox, puis
//         beta  autour de Oy   appliquer  n pour obtenir
//         un vecteur colinaire  z (et orient dans le mme sens)
sommet angles(sommet n, sommet d)
{
  coord alpha,beta,gamma;
  
  // Rotation autour de Ox: (on veut: y=0 et z>=0)
  alpha= pi/2-angle(-n.gety(),n.getz());
  n= rotation(sommet(alpha,0,0),n);
  d= rotation(sommet(alpha,0,0),d);
  // Rotation autour de Oy: (on veut: x=0 et z>=0)
  beta = pi/2-angle(+n.getx(),n.getz());
  d= rotation(sommet(0,beta,0),d);
  // Rotation autour de Oz: (pour que d>0)
  gamma= -angle(-d.getx(),d.gety());
  
  return sommet(alpha,beta,gamma);
}
//-----------------------------------------------------------------------------
void orientation(sommet n1, sommet d1, sommet n2, sommet d2,
			  const ensemble & ES1, const ensemble & ES2)
{
  sommet bary= barycentre(ES1);
  
  sommet a1= angles(n1,d1);
  sommet a2= angles(n2,d2);
  
  rotation(sommet(+a1.getx(),         0,                  0),bary,ES1,ES2);
  rotation(sommet(         0,+a1.gety(),                  0),bary,ES1,ES2);
  rotation(sommet(         0,         0,a1.getz()-a2.getz()),bary,ES1,ES2);
  rotation(sommet(         0,-a2.gety(),                  0),bary,ES1,ES2);
  rotation(sommet(-a2.getx(),         0,                  0),bary,ES1,ES2);
}

void orientation(sommet n1, sommet d1, sommet n2, sommet d2, const ensemble & ES)
{
  orientation(n1,d1,n2,d2,ES,ES);
}
//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------------
// Un sommet isol n'est pas considr comme une face ferme.
bool gcarte::faceFermee(const brin * b0) const
{
  if (b0==NULL)
    return false;
  
  brin * b= (brin *) b0;
  
  do
    b= b->getAlpha(0)->getAlpha(1);
  while (b->getAlpha(1)!=b && b!=b0);
  
  return b->getAlpha(1)!=b;
}
//-----------------------------------------------------------------------------
sommet gcarte::normale(const brin * b0) const
{
  if (!faceFermee(b0))
    return sommet(0,0,1);
  
  const brin * b= b0;
  
  int n=0;
  sommet s(0,0,0);
  
  sommet * S[3];
  S[1]= b->getAlpha(1)->getAlpha(0)->getSommet();
  S[2]= b->getSommet();
  
  do
    {
	 S[0]= S[1];
	 S[1]= S[2];
	 S[2]= b->getAlpha(0)->getSommet();
	 
	 s+= (*S[2] - *S[1]) * (*S[0] - *S[1]);
	 n++;
	 
	 b= b->getAlpha(0)->getAlpha(1);
    }
  while (b!=b0);
  
  if (nul(s.norme()))
    return sommet(0,0,1);
  else
    return s/n;
}
//-----------------------------------------------------------------------------
coord angle(const coord x, const coord y)
{
  if (nul(x))
    {
	 if (y>0)
	   return +pi/2;
	 else
	   return -pi/2;
    }
  else
    {
	 coord a= atan(y/x);
	 if (x<0)
	   a+= pi;
	 
	 if (a>+pi) a-=2*pi;
	 if (a<-pi) a+=2*pi;
	 
	 return a;
    }
}
//-----------------------------------------------------------------------------
