//####################################################################################
// SAUVEGARDE                                                                         
//####################################################################################
#include <qdatetime.h>
#include <qtimer.h>

#define W(s) { if (f.writeBlock((s),QString(s).length())==-1 || f.putch('\n')==-1) return false; }

bool gcarte::sauvegarder(const QString nomfic)
{
  QFile f(nomfic);
  
  f.open(IO_WriteOnly);
  //------------------
  W("; +-----------------------------------------------------------------------+");
  W("; | Fichier x3gc:                                                         |");
  W("; |      Ce fichier est compos de 2 parties:                             |");
  W("; |            - La dfinition des sommets,                               |");
  W("; |            - La dfinition des brins.                                 |");
  W("; |      Les sommets doivent tous tre dclars avant les brins.          |");
  W("; |      Les deux parties doivent tre spares par le caractre '%'.     |");
  W("; |                                                                       |");
  W("; |      1) Dclaration d'un sommet:                                      |");
  W("; |            <nomSommet> ( <x> <y> <z> )                                |");
  W("; |            exemple:                                                   |");
  W("; |               S1       ( -1.3  0.1  2.23 )                            |");
  W("; |                                                                       |");
  W("; |      2) Dclaration d'un brin:                                        |");
  W("; |            <nomBrin> ( <sommet> <alpha0> <alpha1> <alpha2> <alpha3> ) |");
  W("; |            exemple:                                                   |");
  W("; |               B1       ( S13 B4 B7 B1 B1 )                            |");
  W("; |                                                                       |");
  W("; |      Chaque ligne contient la dfinition d'un sommet ou d'un brin.    |");
  W("; |      Des commentaires peuvent tre insrs au moyen du caractre ';'. |");
  W("; |      Tout ce qui suit un ';' sur une ligne est ignor.                |");
  W("; +-----------------------------------------------------------------------+");
  W("");
  
  sommet * s;
  brin * b;
  int i;
  QString ligne;
  
  // On crit la date et l'heure de sauvegarde du fichier.
  QDate date = QDate::currentDate();    // On rcupre la date.
  QTime time = QTime::currentTime();    // On rcupre l'heure.
  char * mois[12] = {"Janvier","Fvrier","Mars","Avril","Mai","Juin","Juillet","Aot","Septembre","Octobre","Novembre","Dcembre"};
  ligne.sprintf("; Sauv le %d %s %d  %dh %dmn %ds\n\n",date.day(),mois[date.month()-1],date.year(), time.hour(),time.minute(),time.second());
  if (f.writeBlock(ligne,ligne.length())==-1) return false;
  
  // Le dictionnaire d'indice de sommets:
  QIntDict<int> ds;
  s= sommet1();
  i= 0;
  while (s!=NULL)
    {
	 ds.insert((unsigned long) s, new int(i++));
	 s= sommetSuivant(s);
    }
  // Le dictionnaire d'indice des brins:
  QIntDict<int> db;
  b= brin1();
  i= 0;
  while (b!=NULL)
    {
	 db.insert((unsigned long) b, new int(i++));
	 b= brinSuivant(b);
    }
  // Les sommets:
  W("; LES SOMMETS:");
  W("; -----------");
  W("");
  s= sommet1();
  while (s!=NULL)
    {
	 ligne.sprintf("S%d\t( % f\t% f\t% f )\n",*ds[(unsigned long) s],s->getx(),s->gety(),s->getz());
	 if (f.writeBlock(ligne,ligne.length())==-1) return false;
	 s= sommetSuivant(s);
    }
  //--------------
  W("\n%\n");
  W("; LES BRINS:    ATTENTION: IL EST RECOMMANDE DE NE PAS MODIFIER CETTE SECTION");
  W("; ---------");
  W("");
  b= brin1();
  while (b!=NULL)
    {
	 ligne.sprintf("B%d\t( S%d\tB%d\tB%d\tB%d\tB%d\t)\n",
				*db[(unsigned long) b],
				*ds[(unsigned long) b->getSommet()],
				*db[(unsigned long) b->getAlpha(0)],
				*db[(unsigned long) b->getAlpha(1)],
				*db[(unsigned long) b->getAlpha(2)],
				*db[(unsigned long) b->getAlpha(3)]);
	 if (f.writeBlock(ligne,ligne.length())==-1) return false;
	 b= brinSuivant(b);
    }
  //--------------
  f.close();
  modif=false;
  return true;
}

#undef W

//###################################################################################
// CHARGEMENT                                                                        
//###################################################################################
QString gcarte::token(QFile * f)
{
  QString s;
  int carlu;
  int etat=0;
  
  while ( (carlu=f->getch()) != -1 )
    {
	 switch (etat)
	   {
	   case 0: // lecture des espaces (tat initial)
		{
		  switch (carlu)
		    {
		    case ' ': case '\t': break;
		    case ';': etat=1; break;
		    case '(': case ')': case '\n': case '%': s.sprintf("%c",carlu); return s;
		    default : etat=2; s.sprintf("%c",carlu); break;
		    }
		  break;
		}
	   case 1: // lecture des commentaires
		{
		  if (carlu=='\n') return QString("\n");
		  break;
		}
	   case 2: // lecture d'un mot:
		{
		  switch (carlu)
		    {
		    case ' ': case '\t': case '(': case ')': case '\n': case '%':
			 f->at(f->at()-1); return s;
		    default: s+= (char) carlu; break;
		    }
		  break;
		}
	   }
    }
  
  return "";
}
//------------------------------------------------------------------------------------
bool gcarte::decimal(QString s)
{
  bool b;
  s.toFloat(&b);
  return b;
}
bool gcarte::id(QString s)
{
  QRegExp re("[a-zA-Z][a-zA-Z0-9_]*",false,true);
  int l;
  
  re.match(s,0,&l);
  return l==s.length();
}
//----------------------------------------------------------------------------------------------------
// retourne 0 si chargement ok, le n de ligne o l'erreur s'est produite sinon
unsigned int gcarte::charger(const QString nomfic, QString & err, bool raz /* remise  zro */)
{
  QDict<sommet>  * dsom = new QDict<sommet>;
  ensemble         esom;
  
  QDict<brin  >  * dbri = new QDict<brin>;
  ensemble         ebri;
  
  QList<QString> * sbr  = new QList<QString>;
  QDict<int    > * lig  = new QDict<int    >;
  
  QFile f(nomfic);
  QString tok;
  QString nom,s,a0,a1,a2,a3; coord x,y,z;
  int n=1;
  int etat;
  
  f.open(IO_ReadOnly);
  // Lecture des sommets:                                                            
  etat=0;
  while (etat!=8)
    {
	 tok= token(&f);
	 
	 switch (etat)
	   {
	   case 0:
		{
		  if (tok=="\n")
		    break;
		  
		  if (tok=="%")
		    { etat=7; break; }
		  
		  if (tok=="")
		    { err.sprintf("Fin de fichier prmature."); return n; }
		  
		  if (!id(tok))
		    { err.sprintf("Nom de sommet incorrect:\n%s",(const char *) tok); return n; }
		  
		  nom=tok;
		  etat=1;
		  break;
		}
	   case 1:
		{
		  if (tok!="(")
		    { err.sprintf("'%s' devrait tre suivi d'une parenthse ouvrante.\n(sur la mme ligne)",(const char *) nom); return n; }
		  
		  etat=2;
		  break;
		}
	   case 2:
		{
		  if (!decimal(tok))
		    { err.sprintf("La valeur '%s' de x est incorrecte.",(const char *) tok); return n; }
		  
		  x=tok.toFloat();
		  etat=3;
		  break;
		}
	   case 3:
		{
		  if (!decimal(tok))
		    { err.sprintf("La valeur '%s' de y est incorrecte.",(const char *) tok); return n; }
		  
		  y=tok.toFloat();
		  etat=4;
		  break;
		}
	   case 4:
		{
		  if (!decimal(tok))
		    { err.sprintf("La valeur '%s' de z est incorrecte.",(const char *) tok); return n; }
		  
		  z=tok.toFloat();
		  etat=5;
		  break;
		}
	   case 5:
		{
		  if (tok!=")")
		    { err.sprintf("Les coordonnes doivent tre suivies d'une parenthse fermante."); return n; }
		  
		  etat=6;
		  break;
		}
	   case 6:
		{
		  if (tok!="\n")
		    { err.sprintf("Rien ne doit suivre la parenthse fermante sur une ligne.\n('%s' trouv)",(const char *) tok); return n; }
		  
		  sommet * ns= new sommet(x,y,z);
		  ns->attache(); // pour la suppression automatique des sommets isols (cas du sommet point par aucun brin)
		  dsom->insert(nom,ns);
		  esom.ajoute(ns);
		  etat=0;
		  break;
		}
	   case 7:
		{
		  if (tok!="\n")
		    { err.sprintf("Le caractre '%%' doit tre seul sur une ligne."); return n; }
		  
		  etat=8;
		  break;
		}
	   }
	 
	 if (tok=="\n") n++;
    }
  // Lecture des brins:                                                              
  etat=0;
  while (etat!=9)
    {
	 tok= token(&f);
	 
	 if (etat>=1 && etat<=7 && tok=="\n")
	   {
		err.sprintf("Fin de ligne atteinte au milieu de la dclaration du brin.");
		return n;
	   }
	 
	 switch (etat)
	   {
	   case 0:
		{
		  if (tok=="")
		    { etat=9; break; }
		  
		  if (tok=="\n")
		    break;
		  
		  if (tok=="(" || tok==')' || tok=="%")
		    {
			 err.sprintf("Nom de brin attendu; caractre '%s' rencontr.",(const char *) tok);
			 return n;
		    }
		  
		  if (!id(tok))
		    {
			 err.sprintf("'%s' n'est pas un nom de brin correct.",(const char *) tok);
			 return n;
		    }
		  
		  if ((*dsom)[tok]!=NULL)
		    {
			 err.sprintf("Le nom '%s' est dj celui d'un sommet.",(const char *) tok);
			 return n;
		    }
		  
		  if ((*dbri)[tok]!=NULL)
		    {
			 err.sprintf("Le brin '%s' a dj t dfini plus haut!!!",(const char *) tok);
			 return n;
		    }
		  
		  nom=tok;
		  etat=1;
		  break;
		}
	   case 1:
		{
		  if (tok!="(")
		    { err.sprintf("'%s' devrait tre suivi sur la mme ligne d'une parenthse ouvrante.",(const char *) nom); return n; }
		  
		  etat=2;
		  break;
		}
	   case 2:
		{
		  if (!id(tok))
		    { err.sprintf("'%s' n'est pas un nom de sommet correct.",(const char *) tok); return n; }
		  
		  if (!(*dsom)[tok])
		    { err.sprintf("'%s' n'est pas un sommet connu.",(const char *) tok); return n; }
		  
		  s=tok;
		  etat=3;
		  break;
		}
	   case 3:
		{
		  if (!id(tok))
		    { err.sprintf("'%s' n'est pas un nom de brin correct.",(const char *) tok); return n; }
		  
		  a0=tok;
		  etat=4;
		  break;
		}
	   case 4:
		{
		  if (!id(tok))
		    { err.sprintf("'%s' n'est pas un nom de brin correct.",(const char *) tok); return n; }
		  
		  a1=tok;
		  etat=5;
		  break;
		}
	   case 5:
		{
		  if (!id(tok))
		    { err.sprintf("'%s' n'est pas un nom de brin correct.",(const char *) tok); return n; }
		  
		  a2=tok;
		  etat=6;
		  break;
		}
	   case 6:
		{
		  if (!id(tok))
		    { err.sprintf("'%s' n'est pas un nom de brin correct.",(const char *) tok); return n; }
		  
		  a3=tok;
		  etat=7;
		  break;
		}
	   case 7:
		{
		  if (tok!=")")
		    { err.sprintf("Caractre ')' attendu; '%s' rencontr.",(const char *) tok); return n; }
		  
		  etat=8;
		  break;
		}
	   case 8:
		{
		  // Mmorisation:
		  brin * b= new brin((*dsom)[s]);
		  
		  dbri->insert(nom,b);
		  ebri.ajoute(b);
		  
		  QString * str= new QString;
		  str->sprintf("%s*%s*%s*%s*%s",
					(const char *) nom,
					(const char *) a0,
					(const char *) a1,
					(const char *) a2,
					(const char *) a3);
		  sbr->append(str);
		  lig->insert(nom,new int(n));
		  // -------------
		  if (tok=="\n") { etat=0; break; }
		  if (tok== "" ) { etat=9; break; }
		  err.sprintf("Fin de ligne attendue; '%s' rencontr.",(const char *) tok);
		  return n;
		}
	   }
	 
	 if (tok=="\n") n++;
    }
  f.close();
  //---------------------------------
  // Vrification de la cohrence:
  QString * ch= sbr->first();
  brin * b;
  
  while (ch!=NULL)
    {
	 int
	   e0= -1,
	   e1= ch->find('*',e0+1),
	   e2= ch->find('*',e1+1),
	   e3= ch->find('*',e2+1),
	   e4= ch->find('*',e3+1),
	   e5= ch->length();
	 
	 nom= ch->mid(e0+1,e1-e0-1);
	 a0 = ch->mid(e1+1,e2-e1-1);
	 a1 = ch->mid(e2+1,e3-e2-1);
	 a2 = ch->mid(e3+1,e4-e3-1);
	 a3 = ch->mid(e4+1,e5-e4-1);
	 
	 b = (*dbri)[nom];
	 b->setAlpha(0,(*dbri)[a0]);
	 b->setAlpha(1,(*dbri)[a1]);
	 b->setAlpha(2,(*dbri)[a2]);
	 b->setAlpha(3,(*dbri)[a3]);
	 
	 ch= sbr->next();
    }
  //---------------------------------
  if (raz)
    nouvelle();

  ES.ajouteEnsemble(esom);
  EB.ajouteEnsemble(ebri);
  //---------------------------------
  // Connexions:
  sommet * sm= sommet1();
  while (sm!=NULL)
    {
	 connect(sm,SIGNAL(isole  (sommet *)),this,SLOT(slotIsole   (sommet *)));
	 connect(sm,SIGNAL(modifie(sommet *)),this,SLOT(slotSModifie(sommet *)));
	 sm= sommetSuivant(sm);
    }
  brin * br= brin1();
  while (br!=NULL)
    {
	 connect(br,SIGNAL(modifie(brin *)),this,SLOT(slotBModifie(brin *)));
	 br= brinSuivant(br);
    }
  // Suppression des sommets isols:
  sm= (sommet *) esom.premier();
  while (sm!=NULL)
    {
	 sm->detache();
	 sm= (sommet *) esom.suivant(sm);
    }
  //---------------------------------
  modif=false;
  emit carteModifiee(this);
  return 0;
}
//-----------------------------------------------------------------------------------------------------
