#include "base.h"
#include "sommet.h"

sommet::sommet()
{
  creation(0,0,0);
}

sommet::sommet(const coord nouvx, const coord nouvy, const coord nouvz)
{
  creation(nouvx,nouvy,nouvz);
}

sommet::sommet(const sommet & s)
{
  creation(s.getx(),s.gety(),s.getz());
}

void sommet::creation(coord X, coord Y, coord Z)
{
  setxyz(X,Y,Z);
  n=0;
}

sommet::~sommet()
{
  if (n!=0)
    erreur("Destruction d'un sommet point par un brin!!!");

  emit modifie(NULL);
}

void sommet::setx(const coord nouvx) { if (x!=nouvx) { x=nouvx; emit modifie(this); } }
void sommet::sety(const coord nouvy) { if (y!=nouvy) { y=nouvy; emit modifie(this); } }
void sommet::setz(const coord nouvz) { if (z!=nouvz) { z=nouvz; emit modifie(this); } }

coord sommet::getx() const { return x; }
coord sommet::gety() const { return y; }
coord sommet::getz() const { return z; }

void sommet::setxyz(const coord nouvx, const coord nouvy, const coord nouvz)
{
  if (x!=nouvx || y!=nouvy || z!=nouvz)
    {
	 x=nouvx;
	 y=nouvy;
	 z=nouvz;
	 emit modifie(this);
    }
}

unsigned int sommet::attache()
{
  return ++n;
}

unsigned int sommet::detache()
{
  if (n==0)
    erreur("Incohrence: Le sommet que vous voulez sparer d'un brin n'est point par aucun brin!");

  if (--n==0)
    emit isole(this);

  return n;
}

unsigned int sommet::nbLiens() const
{
  return n;
}

sommet & sommet::operator= (const sommet & s)
{
  setxyz(s.getx(),s.gety(),s.getz());
  return *this;
}

sommet & sommet::operator+=(const sommet & s)
{
  *this= *this + s;
  return *this;
}

sommet & sommet::operator-=(const sommet & s)
{
  *this= *this - s;
  return *this;
}

sommet & sommet::operator*(const coord k)
{
  return * new sommet(getx()*k,gety()*k,getz()*k);
}

sommet & sommet::operator/(const coord k)
{
  if (k==0)
    erreur("sommet[operator/]: Division par zro.");

  return * new sommet(getx()/k,gety()/k,getz()/k);
}

sommet & sommet::operator*=(const coord k)
{
  return *this = *this * k;
}

sommet & sommet::operator/=(const coord k)
{
  return *this = *this / k;
}

bool sommet::operator==(const sommet & s) const
{
  return
    this->getx() == s.getx() &&
    this->gety() == s.gety() &&
    this->getz() == s.getz();
}

bool sommet::operator!=(const sommet & s) const
{
  return ! (*this == s);
}

coord sommet::norme() const
{
  return sqrt(sqr(getx())+sqr(gety())+sqr(getz()));
}

sommet & sommet::operator+(const sommet & s) const
{
  return * new sommet(getx()+s.getx(), gety()+s.gety(), getz()+s.getz());
}

sommet & sommet::operator-(const sommet & s) const
{
  return * new sommet(getx()-s.getx(), gety()-s.gety(), getz()-s.getz());
}

sommet & sommet::operator*(const sommet & s) const // Produit vectoriel
{
  return * new sommet(gety()*s.getz() - s.gety()*getz(),
				  getz()*s.getx() - s.getz()*getx(),
				  getx()*s.gety() - s.getx()*gety());
}

sommet & sommet::produit(const sommet & a) const // Produit composante par composante
{
  return * new sommet(getx()*a.getx(),
				  gety()*a.gety(),
				  getz()*a.getz());
}

sommet & operator*(const coord k, sommet & s)
{
  return s.operator*(k);
}
