#include "base.h"
#include "superCarte.h"
//-----------------------------------------------------------------------------------
superCarte::superCarte()
{
  pbrin       = NULL;
  epaisseur   = 0;

  update=0;
  cm=ns=false;

  ffermee1    = false;
  ffermee2    = false;

  tout.ajoute(&sbrin1);
  tout.ajoute(&sbrin2);
  tout.ajoute(&sface1);
  tout.ajoute(&sface2);
  tout.ajoute(&svolume1);
  tout.ajoute(&scomposante1);
  tout.ajoute(&sreste);
    
  connect((gcarte *) this,SIGNAL(carteModifiee(gcarte *)),
           		 this,SLOT  (carteModifiee(gcarte *)));
}

superCarte::~superCarte()
{
}
//-----------------------------------------------------------------------------------
void superCarte::majNormales()
{
  brin * b1= (brin *) sbrin1.premier();
  brin * b2= (brin *) sbrin2.premier();

  ffermee1= faceFermee(b1);
  ffermee2= faceFermee(b2);

  if (ffermee1)
    {
	 bary1    = barycentre(sommets(orbiteFace(b1)));
	 normale1 = normale(b1);
	 normale1/= 2*normale1.norme();
    }

  if (ffermee2)
    {
	 bary2    = barycentre(sommets(orbiteFace(b2)));
	 normale2 = normale(b2);
	 normale2/= 2*normale2.norme();
    }
}
//-----------------------------------------------------------------------------------
void superCarte::carteModifiee(gcarte *)
{
  emitCarteModifiee();
}
//-----------------------------------------------------------------------------------
void superCarte::supprimeDoublons(ensemble & e)
{
  ensemble resultat= e;
  ensemble f;

  brin * b = (brin *) e.premier();
  
  while (b!=NULL)
    {
	 if (resultat.contient(b))
	   {
		resultat.enleveEnsemble(orbite(b,false,false,true,true).enleve(b));
	   }

	 b= (brin *) e.suivant(b);
    }

  e = resultat;
}

void superCarte::epure(ensemble & E, ensemble & e)
{
  brin * b= (brin *) e.premier();
  ensemble eArete;

  while (b!=NULL)
    {
	 E.enleveEnsemble(orbite(b,false,false,true,true));

	 b= (brin *) e.suivant(b);
    }
}
//-----------------------------------------------------------------------------------
brin * superCarte::getSelection1()
{
  return (brin *) sbrin1.premier();
}

brin * superCarte::getSelection2()
{
  return (brin *) sbrin2.premier();
}

ensemble superCarte::getSelections()
{
  return sbrin1;
}

brin * superCarte::getBrinPointe()
{
  return pbrin;
}

void superCarte::setBrinPointe(brin * b)
{
  if (b!=pbrin)
    {
	 pbrin= b;
	 emitCarteModifiee();
    }
}
//-----------------------------------------------------------------------------------
void superCarte::addSelection1(brin * b) // Pour l'adjonction de face
{
  if (!sbrin1.estVide() && !orbiteComposante((brin *) sbrin1.premier()).contient(b))
    return;

  ensemble e = orbite(b,false,false,true,true).ajouteEnsemble(orbite(b->getAlpha(0),false,false,true,true));

  if (sbrin1.contient(b))
    sbrin1.enleveEnsemble(e);
  else
    sbrin1.ajouteEnsemble(e);

  sbrin2.vide();

  sreste      = ensembleBrins();

  scomposante1. vide();
  svolume1    . vide();
  sface1      = orbiteComposante(b);
  sface2      . vide();

  // Epuration du reste:
  epure(sreste,sface1);
  supprimeDoublons(sreste);
  // puration face:
  epure(sface1,sbrin1);
  supprimeDoublons(sface1);

  emitCarteModifiee();
}
//-----------------------------------------------------------------------------------
void superCarte::setSelection(brin * b1, brin * b2, bool forceMAJ)
{
  if (!forceMAJ && sbrin1.premier()==b1 && sbrin2.premier()==b2)
    return;

  brin * k;

  beginUpdate();

  sreste      = ensembleBrins();

  scomposante1= orbiteComposante(b1); 
  svolume1    = orbiteVolume    (b1);
  sface1      = orbiteFace      (b1);
  sface2      = orbiteFace      (b2);

  sbrin1.vide(); sbrin1.ajoute(b1);
  sbrin2.vide(); sbrin2.ajoute(b2);

  epaisseur= orbite(b1,false,false,true,true).taille();

  // puration reste:
  epure(sreste,scomposante1);
  epure(sreste,sface2);
  supprimeDoublons(sreste);
  // puration composante:
  epure(scomposante1,svolume1);
  epure(scomposante1,sface2);
  supprimeDoublons(scomposante1);
  // puration volumes:
  epure(svolume1,sface1);
  epure(svolume1,sface2);
  supprimeDoublons(svolume1);
  // puration face:
  epure(sface1,sface2);

  ensemble e1= orbite(b1,false,false,true,true);
  ensemble e2= orbite(b2,false,false,true,true);
  epure(sface1,e1);
  epure(sface1,e2);
  epure(sface2,e1);
  epure(sface2,e2);

  endUpdate();
  emitCarteModifiee();
}

void superCarte::setSelection1(brin * b1, bool forceMAJ)
{
  setSelection(b1,NULL,forceMAJ);
}

void superCarte::setSelection2(brin * b2, bool forceMAJ)
{
  setSelection(getSelection1(),b2,forceMAJ);
}
//-----------------------------------------------------------------------------------
unsigned int superCarte::getEpaisseur() const
{
  return epaisseur;
}
//-----------------------------------------------------------------------------------
const ensemble * superCarte::classesBrins() const
{
  return &tout;
}
//-----------------------------------------------------------------------------------
brin * superCarte::insereFace(int n)
{
  beginUpdate();
  brin * b= gcarte::insereFace(n);

  setSelection1(b,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  endUpdate();
  return b;
}

brin * superCarte::inserePyramide(int n)
{
  beginUpdate();
  brin * b= gcarte::inserePyramide(n);

  setSelection1(b,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  endUpdate();
  return b;
}

brin * superCarte::insereCube()
{
  beginUpdate();
  brin * b= gcarte::insereCube();
  
  setSelection1(b,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  endUpdate();
  return b;
}

brin * superCarte::insereSphere(int m, int p)
{
  beginUpdate();
  brin * b= gcarte::insereSphere(m,p);

  setSelection1(b,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  endUpdate();
  return b;
}

brin * superCarte::insereTore(coord R, coord r, int m, int p)
{
  beginUpdate();
  brin * b= gcarte::insereTore(R,r,m,p);

  setSelection1(b,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  endUpdate();
  return b;
}

brin * superCarte::dupliqueComposante(brin * b)
{
  beginUpdate();
  brin * r = gcarte::dupliqueComposante(b);
  setSelection1(r,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  endUpdate();
  return r;
}

void   superCarte::supprimeComposante(brin * b)
{
  beginUpdate();
  gcarte::supprimeComposante(b);
  setSelection1(NULL,true);
  setBrinPointe(NULL);
  endUpdate();
  emitNouvelleSelection();
}

bool superCarte::coudre(brin * a,brin * b, bool o, bool t, bool p)
{
  bool resultat= gcarte::coudre(a,b,o,t,p);
  setSelection1(a,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  return resultat;
}

brin * superCarte::decoudre(brin * b)
{
  brin * nb = gcarte::decoudre(b);
  setSelection1(nb,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  return nb;
}

brin * superCarte::extrusionFace(brin * b, sommet v, unsigned int m)
{
  brin * nb= gcarte::extrusionFace(b,v,m);
  setSelection1(nb,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  return nb;
}

brin * superCarte::extrusionMultiple(brin * f, brin * c)
{
  brin * nb= gcarte::extrusionMultiple(f,c);
  setSelection1(nb,true);
  setBrinPointe(NULL);
  emitNouvelleSelection();
  return nb;
}
//-----------------------------------------------------------------------------------
void superCarte::nouvelle()
{
  beginUpdate();
  gcarte::nouvelle();
  setSelection(NULL,NULL,true);
  setBrinPointe(NULL);
  emitCarteModifiee();
  endUpdate();
}

unsigned int superCarte::charger(QString nomfic, QString & message, bool raz)
{
  beginUpdate();
  unsigned int v= gcarte::charger(nomfic,message,raz);

  if (v==0)
    {
	 setSelection(NULL,NULL,true);
	 setBrinPointe(NULL);
    }

  endUpdate();
  return v;
}
//-----------------------------------------------------------------------------------
void superCarte::ajouteSommet(brin * b, sommet s)
{
  beginUpdate();
  gcarte::ajouteSommet(b,s);
  setSelection1(b,true);
  setBrinPointe(NULL);
  endUpdate();
}

void superCarte::ajouteArete(brin * b1, brin * b2)
{
  beginUpdate();
  gcarte::ajouteArete(b1,b2);
  setSelection1(b1,true);
  setBrinPointe(NULL);
  endUpdate();
}

brin * superCarte::ajouteFace(ensemble e)
{
  beginUpdate();
  brin * resultat= gcarte::ajouteFace(e);
  setSelection1(resultat,true);
  setBrinPointe(NULL);
  endUpdate();
  return resultat;
}
//-----------------------------------------------------------------------------------
brin * superCarte::coutureParBord(brin * b1, brin * b2)
{
  beginUpdate();
  brin * resultat= gcarte::coutureParBord(b1,b2);
  setSelection1(resultat,true);
  setBrinPointe(NULL);
  endUpdate();
  return resultat;
}
//-----------------------------------------------------------------------------------
brin * superCarte::fusion(brin * b)
{
  beginUpdate();
  brin * resultat= gcarte::fusion(b);
  setSelection1(resultat,true);
  setBrinPointe(NULL);
  endUpdate();
  return resultat;
}
//-----------------------------------------------------------------------------------
brin * superCarte::ajouteBrin(sommet * s)
{
  beginUpdate();
  brin * resultat= gcarte::ajouteBrin(s);
  endUpdate();
  return resultat;
}

void   superCarte::supprimeBrin(brin * b)
{
  beginUpdate();
  gcarte::supprimeBrin(b);
  endUpdate();
}
//-----------------------------------------------------------------------------------
bool   superCarte::faceFermee1    () const { return ffermee1; }
bool   superCarte::faceFermee2    () const { return ffermee2; }
sommet superCarte::barycentreFace1() const { if (ffermee1) return bary1   ; else return sommet(0,0,0); }
sommet superCarte::barycentreFace2() const { if (ffermee2) return bary2   ; else return sommet(0,0,0); }
sommet superCarte::normaleFace1   () const { if (ffermee1) return normale1; else return sommet(0,0,0); }
sommet superCarte::normaleFace2   () const { if (ffermee2) return normale2; else return sommet(0,0,0); }
//-----------------------------------------------------------------------------------
void superCarte::emitCarteModifiee()
{
  if (update==0)
    {
	 majNormales();
	 emit carteModifiee(this);
    }
  else
    cm=true;
}

void superCarte::emitNouvelleSelection()
{
  if (update==0)
    {
	 majNormales();
	 emit nouvelleSelection(this);
    }
  else
    ns=true;
}
//-----------------------------------------------------------------------------------
void superCarte::beginUpdate()
{
  update++;
}

void superCarte::endUpdate()
{
  update--;
  
  if (update<0)
    erreur("superCarte[endUpdate]: Appel de endUpdate avant beginUpdate...");

  if (update==0 && (cm || ns))
    {
	 majNormales();
	 if (cm) emit carteModifiee    (this);
	 if (ns) emit nouvelleSelection(this);

	 cm=ns=false;
    }
}
//-----------------------------------------------------------------------------------
